from __future__ import print_function
from . import debug
from . import raw_mode
from infi.shell.rlcontext import get_active_rlcontext
from infi.shell.shell import get_active_shell
from infi.shell.config import config
from infi.shell.utils import Timer
from .exceptions import (
    ConnectionFailed,
    BadURI,
    Unauthorized,
    ServiceTemporarilyUnavailable,
    InternalServerError,
    ConfigurationConflict,
    NotSupported,
    Forbidden,
    UnexpectedServerResponse,
    ManagementException,
    BadPageStructure,
    )
from infi.shell.exceptions import CommandAborted, NoActiveReadlineContext
from . import VERSION as INFINISHELL_VERSION
from bunch import bunchify, Bunch
import json
import re
from colorama import Style, Fore
import requests
from urllib2 import quote
import httplib
from itertools import chain
from functools import partial
from inspect import isroutine
from contextlib import contextmanager

_REQUEST_TITLE = Style.BRIGHT + Fore.MAGENTA
_RESPONSE_TITLE = Style.BRIGHT + Fore.RED
_REQUEST_HEADERS = Fore.MAGENTA
_RESPONSE_HEADERS = Fore.RED
_TIME_SUMMARY = Style.BRIGHT + Fore.YELLOW
_ENDTITLE = Style.RESET_ALL
_EOL = Style.RESET_ALL
NOTHING = object()

def _print_debug(msg, *args, **kwargs):
    if debug.is_on() or raw_mode.is_on():
        if args or kwargs:
            msg = msg.format(*args, **kwargs)
        print(msg)
def _format_headers(headers, color):
    max_length = max(len(k) for k in headers)
    template = "{}{:<" + str(max_length) + "} : {}{}{}"
    return "    " + "\n    ".join(template.format(color, k, Style.BRIGHT, v, _EOL) for k, v in headers.iteritems())

class RestAPI(object):
    def __init__(self, baseuri, headers={}):
        self._baseuri = baseuri.rstrip("/") + "/"
        _headers = self._get_default_headers()
        _headers.update(headers)
        self.set_username(None)
        self.session = self._create_session(_headers)
    def _create_session(self, headers):
        session = requests.session()
        session.headers.update(headers)
        return session
    def set_basic_auth(self, username, password):
        self.session.auth = (username, password)
        self.set_username(username)
    @contextmanager
    def change_basic_auth_context(self, username, password):
        old_auth = self.session.auth
        self.set_basic_auth(username, password)
        try:
            yield
        except:
            self.set_basic_auth(*old_auth)
            raise
    @contextmanager
    def no_basic_auth_context(self):
        auth, self.session.auth = self.session.auth, None
        try:
            yield
        finally:
            self.session.auth = auth
    def get_username(self):
        return self.username
    def set_username(self, username):
        self.username = username
    @classmethod
    def _get_default_headers(cls):
        return {}
    def _pre_send(self, data):
        return data if data is not NOTHING else None
    def _send(self, uri, method, data=NOTHING, headers={}, params=None):
        data = self._pre_send(data)
        full_uri = self._baseuri + uri
        response = self._send_request(method, full_uri, data, headers, params)
        self._post_receive(response)
        return response
    def _send_request(self, method, uri, data, headers, params):
        if debug.is_on():
            full_uri = uri
            if params:
                full_uri = "{}{}{}".format(uri, "&" if "?" in uri else "?",
                                           "&".join("{}={}".format(quote(str(key)), quote(str(value)))
                                                    for key, value in params.iteritems()))
            _print_debug(_REQUEST_TITLE + u"### REQUEST: " + _ENDTITLE +
                         Style.BRIGHT + Fore.CYAN + method + " " + Fore.BLUE + full_uri + _EOL)
            all_headers = dict(chain(self.session.headers.iteritems(), headers.iteritems()))
            self._print_headers(all_headers, _REQUEST_HEADERS)
            self._print_data(data, _REQUEST_HEADERS)
        try:
            timer = Timer()
            with timer:
                response_obj = self.session.request(
                    method,
                    uri,
                    params=params,
                    data=data,
                    headers=headers,
                    timeout=config.api_request_timeout_seconds,
                    verify=config.api_ssl_verify_certificate,
                    )
            response = Bunch(
                code=response_obj.status_code,
                data=response_obj.text,
                headers=response_obj.headers,
                response_object=response_obj,
                )
        except requests.Timeout as e:
            raise ConnectionFailed("Connection attempt timed out.")
        except requests.ConnectionError as e:
            try:
                error_message = e.args[0].reason.strerror.capitalize()
            except:
                error_message = str(e)
            raise ConnectionFailed(error_message)
        if debug.is_on():
            _print_debug(_RESPONSE_TITLE + "### RESPONSE: " + _ENDTITLE + Style.BRIGHT + Fore.CYAN +
                         "{} {}".format(response.code, httplib.responses[response.code]) + _EOL)
            self._print_headers(response.headers, _RESPONSE_HEADERS)
            self._print_data(response.data, _RESPONSE_HEADERS)
            _print_debug(_TIME_SUMMARY + "Total time: {:.2} seconds".format(timer.get_seconds()) + _EOL)
            _print_debug("")
        elif raw_mode.is_on():
            shell = get_active_shell(safe=True)
            if shell and shell.state.is_during_command():
                _print_debug(Fore.BLUE + uri + _EOL)
                self._print_data(response.data, _RESPONSE_HEADERS, with_title=False)
        return response
    def _print_headers(self, headers, color):
        _print_debug(_format_headers(headers, color))
    def _print_data(self, data, color, with_title=True):
        if data:
            _print_debug(color + ("DATA: " if with_title else "") + Fore.GREEN + Style.BRIGHT + data + _EOL)
        else:
            _print_debug(color + "NO DATA" + _EOL)
    def _post_receive(self, response):
        return
    def get(self, uri, params={}):
        return self._send(uri, "GET", params=params)
    def post(self, uri, data=NOTHING):
        return self._send(uri, "POST", data)
    def put(self, uri, data=NOTHING):
        return self._send(uri, "PUT", data)
    def delete(self, uri):
        return self._send(uri, "DELETE")

class JsonRestAPI(RestAPI):
    _acceptable_codes = 200, 201, 204
    _confirmation_code = 403
    _HEADERS = Bunch(
        CONFIRMATION = "x-infinidat-approved",
        CLIENT_ID = "x-infinidat-source",
        WARNINGS = "x-infinidat-warnings",
        )
    _API_WARNING_SEPARATOR = ";"
    _MGMT_CODES = Bunch(
        APPROVAL = "APPROVAL_REQUIRED",
        )
    def __init__(self, baseuri, client_id, **kwargs):
        # Set client ID header
        kwargs.setdefault("headers", {})[self._HEADERS.CLIENT_ID] = client_id
        super(JsonRestAPI, self).__init__(baseuri, **kwargs)
        self._error_handler = ErrorHandler()
        self._api_warnings = set()
    @classmethod
    def _get_default_headers(cls):
        headers = super(JsonRestAPI, cls)._get_default_headers()
        headers.update({"content-type": "application/json",
                        "accept-charset": "utf-8",
                        "user-agent": "InfiniShell/{}".format(INFINISHELL_VERSION)})
        return headers
    def _send(self, *args, **kwargs):
        response = super(JsonRestAPI, self)._send(*args, **kwargs)
        # Command may fail if it's considered "destructive"
        if self._is_confirmation_required(response):
            # Confirm interactively and send again including confirmation header
            if not config.auto_confirm_prompts:
                confirmation_prompt = None if not isinstance(response.data, dict) else response.data.get('message')
                if not self._get_user_confirmation(confirmation_prompt):
                    raise CommandAborted
            kwargs.setdefault("headers", {})[self._HEADERS.CONFIRMATION] = "yes"
            response = super(JsonRestAPI, self)._send(*args, **kwargs)
        if response.code not in self._acceptable_codes:
            self._error_handler.handle(response.code, response.data)
        return response.data
    def _get_management_code(self, response):
        if (not isinstance(response.data, dict)
            or "code" not in response.data
            or not isinstance(response.data["code"], basestring)):
            return None
        return response.data["code"].upper()
    def _is_confirmation_required(self, response):
        return response.code == self._confirmation_code and self._get_management_code(response) == self._MGMT_CODES.APPROVAL
    def _get_user_confirmation(self, msg=None):
        try:
            if msg:
                msg = msg.strip()
                if msg[-1] not in ['.', '?', '!']:
                    msg += '.'
            else:
                msg = 'This action requires your confirmation.'
            while True:
                answer = get_active_rlcontext().input(prompt="{}\nAre you sure? [y/n] ".format(msg),
                                                      auto_completion=False,
                                                      history=False).lower()
                if answer in ("y", "yes"): return True
                if answer in ("n", "no"): return False
        except NoActiveReadlineContext:
            raise CommandAborted("Command requires user confirmation (use -y or the interactive shell)")
    def _pre_send(self, data):
        return json.dumps(data, indent=2) if data is not NOTHING else None
    def _post_receive(self, response):
        super(JsonRestAPI, self)._post_receive(response)
        for attr in (a for a in dir(self) if a.startswith("_post_receive__")):
            method = getattr(self, attr)
            if isroutine(method):
                method(response)
    def _post_receive__body(self, response):
        if response.headers.get("content-type") == "application/json":
            try:
                json_data = json.loads(response.data)
            except ValueError:
                return
            response.data = bunchify(json_data)
    def _post_receive__api_warnings(self, response):
        shell = get_active_shell(safe=True)
        # No point in doing this unless shell is active, and only during
        # command execution, i.e not during auto-completion and such
        if not shell or not shell.state.is_during_command():
            return
        api_warnings = response.headers.get(self._HEADERS.WARNINGS)
        if api_warnings:
            for warning in api_warnings.split(self._API_WARNING_SEPARATOR):
                warning = warning.strip()
                if warning and warning not in self._api_warnings:
                    self._api_warnings.add(warning)
                    shell.output("API warning: " + warning, colors=[Style.BRIGHT, Fore.YELLOW], to_stderr=True)
    ##### Infinibox logic #####
    def paged_query(self, uri, page_size=None, params=None, hook=None):
        return PagedQuery(self, uri, page_size or config.api_default_page_size, params=params, hook=hook)
    def get_sub_resource(self, uri, **extra_attrs):
        return APISubResource.create(self, uri, **extra_attrs)

class APISubResource(object):
    def __init__(self, api, base_uri):
        self.__api__ = api
        self.__base_uri__ = base_uri.rstrip("/") + "/"
    def __get_full_uri__(self, uri):
        return self.__base_uri__ + uri.lstrip("/")
    def get(self, uri, *args, **kwargs):
        return self.__api__.get(self.__get_full_uri__(uri), *args, **kwargs)
    def post(self, uri, *args, **kwargs):
        return self.__api__.post(self.__get_full_uri__(uri), *args, **kwargs)
    def put(self, uri, *args, **kwargs):
        return self.__api__.put(self.__get_full_uri__(uri), *args, **kwargs)
    def delete(self, uri, *args, **kwargs):
        return self.__api__.delete(self.__get_full_uri__(uri), *args, **kwargs)
    def paged_query(self, uri, *args, **kwargs):
        return self.__api__.paged_query(self.__get_full_uri__(uri), *args, **kwargs)
    def __getattr__(self, attr):
        return getattr(self.__api__, attr)
    def __setattr__(self, attr, value):
        if attr not in ("__api__", "__base_uri__"):
            raise TypeError("Cannot set additional attributes in {}".format(self.__class__.__name__))
        return object.__setattr__(self, attr, value)
    @classmethod
    def create(cls, api, base_uri, **extra_attrs):
        class proxy_class(cls):
            pass
        proxy_class.__name__ = cls.__name__
        for attr, value in extra_attrs.iteritems():
            setattr(proxy_class, attr, value)
        return proxy_class(api, base_uri)

class ErrorHandler(object):
    _exceptions = {
        401 : Unauthorized,
        403 : Forbidden,
        404 : BadURI,
        409 : ConfigurationConflict,
        500 : InternalServerError,
        501 : NotSupported,
        503 : ServiceTemporarilyUnavailable,
        }
    _HTML_PATTERN = re.compile("<html>", re.IGNORECASE)
    def handle(self, code, data):
        msg = None
        if data:
            if isinstance(data, dict):
                if "message" in data:
                    if "code" in data:
                        raise ManagementException(data["code"], data["message"], code)
                    msg = data["message"]
            elif not isinstance(data, basestring):
                msg = data
            elif not self._HTML_PATTERN.match(data):
                msg = None
        AppropriateException = self._exceptions.get(code) or partial(UnexpectedServerResponse, code)
        raise AppropriateException(msg) if msg is not None else AppropriateException()

class PagedQuery(object):
    _PAGE_FIELDS = ("objects", "page_info", "metadata")
    def _assert_params(self, params):
        for param in ("page", "page_size"):
            assert param not in params, "'{}' is a reserved parameter, and cannot be specified in 'params'".format(param)
        return True
    def __init__(self, api, url, page_size=None, params=None, hook=None):
        params = params or {}
        assert self._assert_params(params)
        self.api = api
        self.url = url
        self._objects_hook = hook
        self._params = Bunch(params)
        if page_size is not None:
            self._params.page_size = page_size
        self._first_page = self._get_page(1)
        self.page_size = self._first_page.page_info.page_size
        self.total_pages = self._first_page.page_info.pages_total
        self.total_objects = self._first_page.metadata.number_of_objects
        self.changelog_id = self._first_page.metadata.change_log_id
        # Update params for next pages
        self._params.page_size = self.page_size
        #self._params.change_log_id = self.changelog_id
    def __iter__(self):
        return self.iter_all_objects()
    def iter_pages(self):
        for number in xrange(1, self.total_pages+1):
            yield self.get_page(number)
    def _get_page(self, number):
        self._params.page = number
        page_data = self.api.get(self.url, params=self._params)
        self._assert_page_data(page_data)
        if self._objects_hook:
            page_data.objects = self._objects_hook(page_data.objects)
        return page_data
    def _assert_page_data(self, page_data):
        if not isinstance(page_data, dict) or not all(f in page_data for f in self._PAGE_FIELDS):
            raise BadPageStructure(page_data)
    def get_page(self, number):
        assert number >= 1, "Page numbers are 1-based"
        page_data = self._first_page if number == 1 else self._get_page(number)
        return Page(self, number, page_data.objects)
    def get_all_objects(self):
        return list(self.iter_all_objects())
    def iter_all_objects(self):
        for page in self.iter_pages():
            for object in page:
                yield object

class Page(object):
    def __init__(self, query, number, objects):
        self.query = query
        self.number = number
        self.objects = objects
    def __iter__(self):
        return iter(self.objects)
