from ..renderer import CollectionRenderer
from ..exceptions import UnexpectedAPIResponse
from infi.shell.commands.commands import invalidates_autocompletion_cache, availability, render
from ..system import system as current_system # There's a module named "system"
from ..system.user import Role
from functools import partial

Null, MESSAGE, ENTITY = object(), object(), object()
#MAX_POSSIBLE_VALUES = 200

class EntityCommands(object):
    ENTITY_TYPE = None
    def __init__(self):
        assert self.ENTITY_TYPE
    def _get_actual_value(self, v):
        if v is Null:
            return None
        if isinstance(v, dict) and "id" in v:
            return v["id"]
        return v
    def _iter_properties(self, properties):
        """ Properties which value is None are ignored, and
        properties which value is Null are assigned None """
        return ((k, self._get_actual_value(v))
                for k, v in properties.iteritems() if v is not None)
    def _filter_properties(self, properties):
        """ The same as _iter_properties(), but returns a dict """
        return dict(self._iter_properties(properties))
    def _query(self, _api, _entity=None, _hook=None, **conditions):
        """ Get iterator for all entities matching the specified conditions. If
        the first positional argument is used, then the iterator will return
        it, providing it matches the specified conditions. """
        conditions = self._filter_properties(conditions)
        if _entity:
            # Entity was specified - use it
            results = [_entity]
            if _hook:
                results = _hook(results)
        elif conditions.get(self.ENTITY_TYPE.NAME_FIELD, None):
            # Get by name, only one entity is expected (or exception)
            results = [self.ENTITY_TYPE.get_by_name(_api, conditions.pop(self.ENTITY_TYPE.NAME_FIELD))]
        else:
            if self.ENTITY_TYPE.PAGED:
                hook = None if not conditions else partial(self._matches_query_hook, conditions)
                if hook is None:
                    hook = _hook
                elif _hook is not None: # Two hooks
                    matches_hook = hook
                    hook = lambda results: _hook(matches_hook(results))
                return self.ENTITY_TYPE.paged_query(_api, hook=hook)
            else:
                results = self.ENTITY_TYPE.iter_all(_api)
        if not conditions:
            return results
        return self._iterate_matching(results, conditions)
    def _matches_query_hook(self, conditions, results):
        return list(self._iterate_matching(results, conditions))
    def _paged_query(self, _api, *args, **kwargs):
        """ Get a PagedQuery object for this group's entity type. """
        assert self.ENTITY_TYPE.PAGED, '{} does not support paging'.format(self.ENTITY_TYPE.get_entity_name())
        return self.ENTITY_TYPE.paged_query(_api, *args, **kwargs)
    def _iterate_matching(self, entities, conditions):
        return (entity for entity in entities if self._matches_query(entity, conditions))
    def _matches_query(self, entity, conditions):
        for property, value in conditions.iteritems():
            if not hasattr(entity, property) or not value == getattr(entity, property):
                return False
        return True
    def _result_message(self, entity, msg, subject=None):
        return '{} "{}" {}'.format((subject or self.ENTITY_TYPE.get_entity_name()).capitalize(), entity[self.ENTITY_TYPE.NAME_FIELD], msg)
    @invalidates_autocompletion_cache
    def _create(self, _api, returns=MESSAGE, **properties):
        """ Create an entity and return an appropriate message. """
        entity = _api.post(self.ENTITY_TYPE.get_uri(), self._filter_properties(properties))
        if returns is ENTITY:
            return entity
        if returns is MESSAGE:
            return self._result_message(entity, "created")
        return returns
    def _create_and_return_entity(self, _api, **properties):
        return self._create(_api, returns=ENTITY, **properties)
    def _update(self, _api, entity, returns=MESSAGE, **properties):
        """ Update an entity and return an appropriate message. """
        for property, value in self._iter_properties(properties):
            _api.put("{}/{}/{}".format(self.ENTITY_TYPE.get_uri(), entity[self.ENTITY_TYPE.ID_FIELD], property), value)
        if returns is ENTITY:
            return entity
        if returns is MESSAGE:
            return self._result_message(entity, "updated")
        return returns
    @invalidates_autocompletion_cache
    def _rename(self, _api, entity, new_name):
        """ Rename an entity and return a user-friendly message. """
        self._update(_api, entity, returns=ENTITY, name=new_name)
        return self._result_message(entity, 'renamed to "{}"'.format(new_name))
    @invalidates_autocompletion_cache
    def _delete(self, _api, entity):
        """ Delete an entity and return a user-friendly message. """
        deleted = _api.delete("{}/{}".format(self.ENTITY_TYPE.get_uri(), entity[self.ENTITY_TYPE.ID_FIELD]))
        if not isinstance(deleted, dict):
            raise UnexpectedAPIResponse(u"Deletion return value expected to be a {} object, but instead was: {}"
                                        .format(self.ENTITY_TYPE.get_entity_name(), deleted))
        if entity[self.ENTITY_TYPE.ID_FIELD] != deleted[self.ENTITY_TYPE.ID_FIELD]:
            raise UnexpectedAPIResponse("Deleted ID differs from specified ID: {}, {}".format(deleted[self.ENTITY_TYPE.ID_FIELD], entity[self.ENTITY_TYPE.ID_FIELD]))
        return self._result_message(entity, "deleted")

def _available_to_role_predicate(role):
    return current_system.user.role.is_role(role)

def available_to_role(func=None, role=None):
    """ Command method availability decorator which filters commands based on
    user roles. This decorator may be used multiple times on the same method.
    If not used, method will be available to all user roles."""
    assert role, 'Role must be specified'
    return availability(func, predicate=partial(_available_to_role_predicate, role))

def superuser(func=None):
    return available_to_role(func, Role.INFINIDAT)

def admin_only(func=None):
    return available_to_role(func, Role.ADMIN)

_collection_renderer = CollectionRenderer()
def render_collection(func=None, **kwargs):
    return render(func=func, renderer=_collection_renderer, **kwargs)
