from . import EntityCommands, render_collection
from .types import EventCode, EventLevel, EventReporter, EventVisibility, EventRule, Event, Name
from ..system import system
from ..exceptions import NoSuchEntity
from infi.shell.commands import commandgroup, command, render
from infi.shell.commands.types import Timestamp, Str, StrList, Bool, Int
from infi.shell.exceptions import CommandFailed, BadValue
from bunch import Bunch
from collections import OrderedDict

## WORKAROUND ##
from ..renderer import InfiniboxRenderer
from itertools import chain
import sys
class DumpRenderer(InfiniboxRenderer):
    def _render(self, events):
        events = iter(events)
        self._stream = sys.stdout
        try:
            first_event = next(events)
        except StopIteration:
            return
        fields, rendered_field_names, excluded_fields = self._get_fields(first_event)
        events = self._iter_and_render(chain([first_event], events), excluded_fields)
        self._render_table_rows(events, fields)
    def _iter_and_render(self, objects, excluded_fields):
        for o in objects:
            for field, value in o.items():
                if field in excluded_fields:
                    continue
                # Render value
                rendered = self._render_value(o, field, value)
                if value is not rendered:
                    o[field] = rendered
                    value = rendered
            yield o
    def _join_strings(self, objects):
        return u",".join(str(o) for o in objects)
    def _render_table_rows(self, objects, fields):
        self._writeln(self._join_strings(fields))
        for o in objects:
            # Format values, while keeping fields order
            self._writeln(self._join_strings(o.get(f, "") for f in fields))
DUMP_RENDERER = DumpRenderer()
## WORKAROUND ##

@commandgroup(prefix="event")
class EventCommands(EntityCommands):
    ENTITY_TYPE = Event

    @command(types=(Timestamp, Timestamp, Int, Int, EventCode, EventLevel),
             substitute=dict(from_time="from", to_time="to"))
    @render_collection(entity=Event, alignment="left", forced=["id"],
                       order=["timestamp", "id", "source_node_id", "level", "reporter", "code"],
                       excluded=["visibility", "description_template", "system_version", "affected_entity_id", "data"],
                       substitute=dict(source_node_id="node"))
    def query(self, _api, from_time=None, to_time=None, from_id=None, to_id=None, code=None, level=None):
        return self._event_query(_api, from_time, to_time, from_id, to_id, code, level, page_size=1000, hook=self._event_query_hook)

    @command(types=(Timestamp, Timestamp, Int, Int, EventCode, EventLevel),
             substitute=dict(from_time="from", to_time="to"))
    @render(renderer=DUMP_RENDERER,
            order=["timestamp", "id", "level", "reporter", "code"], forced=["id"],
            excluded=["visibility", "description_template"], alignment="left")
    def dump(self, _api, from_time=None, to_time=None, from_id=None, to_id=None, code=None, level=None):
        ''' This is a workaround command until paged queries are properly
        supported by the renderer. '''
        return self._event_query(_api, from_time, to_time, from_id, to_id, code, level, page_size=1000, hook=self._event_dump_hook)

    def _event_query(self, _api, from_time, to_time, from_id, to_id, code, level, page_size=None, hook=None):
        if None not in (from_time, to_time) and from_time > to_time:
            raise BadValue("'from' timestamp must not be higher than 'to' timestamp")
        filter = []
        if from_time is not None:
            filter.append("timestamp^>=^{}".format(from_time))
        if to_time is not None:
            filter.append("timestamp^<=^{}".format(to_time))
        if from_id is not None:
            filter.append("id^>=^{}".format(from_id))
        if to_id is not None:
            filter.append("id^<=^{}".format(to_id))
        if code is not None:
            filter.append("code^IN^({})".format(code.upper()))
        if level is not None:
            filter.append("level^IN^({})".format(level.upper()))
        params = dict(sort="timestamp^DESC")
        if filter:
            params["filter"] = "^AND^".join(filter)
        return self._paged_query(_api, params=params, hook=hook, page_size=page_size)
    def _event_query_hook(self, events):
        for e in events:
            if e.description:
                e.description = e.description.replace("\n", " ")
        return events
    def _event_dump_hook(self, events):
        for e in events:
            if e.description:
                e.description = e.description.replace("\n", " ")
            if e.data:
                data = sorted((field.name, field.value.replace("\n", " ")) for field in e.data
                              # Show only non-strings, and non-empty strings (omit empty strings)
                              if not isinstance(field.value, basestring) or field.value
                              # Hide SysConf data (in e.g heartbeat events)
                              and field.type.lower() != "sysconf")
                e.data = " ".join("{}={}".format(key, value.replace("\n", " "))
                                  for key, value in data)
        return events

    @command(types=(Int,))
    @render(individually=True, excluded=["visibility", "description_template"], forced=["id"])
    def details(self, _api, id):
        event = Event.get_by_id(_api, id)
        data = event.data
        event = OrderedDict(sorted((key, value) for key, value in event.iteritems() if key not in ("visibility", "description_template", "data")))
        for param in sorted(data, key=lambda p: p.name):
            event['data: ' + param.name] = param.value
        return event
    def _format_constants(self, _api, clazz, column_name):
        return [Bunch({column_name:value}) for value in clazz.iter_all(_api)]
    @command
    def codes(self, _api):
        return self._format_constants(_api, EventCode, 'code')
    @command
    def levels(self, _api):
        return self._format_constants(_api, EventLevel, 'level')
    @command
    def reporters(self, _api):
        return self._format_constants(_api, EventReporter, 'reporter')
    @command
    def visibilities(self, _api):
        return self._format_constants(_api, EventVisibility, 'visibility')
    @command(types=(EventLevel, Str))
    def create(self, _api, level, description):
        _api.post("events/custom", dict(
            visibility="CUSTOMER",
            level=level,
            description_template=description,
            ))
        return "Event created."

@commandgroup(prefix="event")
class EventRuleCommands(EntityCommands):
    ENTITY_TYPE = EventRule
    @command(types=(EventRule,),
             substitute=dict(rule="name"))
    @render(alignment="left",
            order=["field_name", "is_data", "accepted_values",
                   "name", "visibility", "filters", "recipients"])
    def rule_query(self, _api, rule=None):
        if not rule:
            return self._query(_api, _hook=self._rule_query_hook)
        return rule.filters
    def _rule_query_hook(self, rules):
        for rule in rules:
            rule.filters = [filter.field_name for filter in rule.filters]
        return rules
    @command(types=(Name, StrList))
    def rule_create(self, _api, name, recipients):
        return self._create(_api,
                name=name,
                recipients=recipients,
                filters=[],
                visibility="CUSTOMER" if system.user.role.is_admin() else "INFINIDAT",
                )
    @command(types=(EventRule, Name, StrList),
             substitute=dict(rule="name"))
    def rule_update(self, _api, rule, new_name=None, recipients=None):
        if (new_name, recipients).count(None) == 2:
            raise CommandFailed("Either name, recipients, or both, must be specified")
        return self._update(_api, rule, name=new_name, recipients=recipients)
    @command(types=(EventRule,),
             substitute=dict(rule="name"))
    def rule_delete(self, _api, rule):
        return self._delete(_api, rule)
    @command(types=(EventRule, Str, Bool, StrList),
             substitute=dict(rule="name"))
    def rule_add_filter(self, _api, rule, field, is_data, values):
        _api.post("{}/{}/filters".format(self.ENTITY_TYPE.get_uri(), rule.id), dict(
            field_name=field,
            is_data=is_data,
            accepted_values=values,
            ))
        return "Filter added"
    @command(types=(EventRule, Str),
             substitute=dict(rule="name"))
    def rule_remove_filter(self, _api, rule, field):
        for filter in rule.filters:
            if filter.field_name == field:
                _api.delete("{}/{}/filters/{}".format(self.ENTITY_TYPE.get_uri(), rule.id, filter.id))
                return "Filter removed"
        raise NoSuchEntity('Rule "{}" has no such filter: {}'.format(rule.name, field))
