from infi.shell.commands import commandgroup, command, render
from infi.shell.commands.types import Int
from infi.shell.exceptions import BadArguments, BadValue
from ..exceptions import BadURI, UnexpectedAPIResponse
from .types import Name, Host, Port, PortType, Volume
from . import EntityCommands, render_collection
from bunch import Bunch

class HostCommandsBase(EntityCommands):
    ENTITY_TYPE = Host
    def _assert_result(self, expected, actual, subject="value"):
        if expected != actual:
            raise UnexpectedAPIResponse(u"Returned {0} differs from specified {0}.\nExpected: {1}\nGot: {2}"
                                        .format(subject.replace("_", " "), expected, actual))
    def _map(self, _api, host_like, volume, lun):
        props = {"volume_id" : volume.id}
        if lun is not None:
            props["lun"] = lun
        result = _api.post("{}/{}/luns".format(self.ENTITY_TYPE.get_uri(), host_like.id), props)
        self._assert_result(volume.id, result["volume_id"], "ID")
        if lun is not None:
            self._assert_result(lun, result.lun, "LUN")
        if "lun" in result:
            lun_text = "LUN {} in ".format(result.lun)  # noqa
        return 'Volume "{volume.name}" mapped to {lun_text}{ENTITY_TYPE} "{host_like.name}"'\
               .format(ENTITY_TYPE=self.ENTITY_TYPE.get_entity_name(), **locals())
    def _unmap(self, _api, host_like, volume, lun):
        if (volume, lun).count(None) != 1:
            raise BadArguments("Either volume or lun must be specified")
        try:
            if lun is not None:
                url = "{}/{}/luns/lun/{}".format(self.ENTITY_TYPE.get_uri(), host_like.id, lun)
            elif volume is not None:
                url = "{}/{}/luns/volume_id/{}".format(self.ENTITY_TYPE.get_uri(), host_like.id, volume.id)
            result = _api.delete(url)
        except BadURI:
            raise BadValue('LUN {} is not mapped in {} "{}"'.format(lun, self.ENTITY_TYPE.get_entity_name(), host_like.name))
        volume_name = volume.name if volume else self._get_volume_name(_api, result.volume_id)  # noqa
        return 'Volume "{volume_name}" unmapped from LUN {result.lun} in {ENTITY_TYPE} "{host_like.name}"'\
               .format(ENTITY_TYPE=self.ENTITY_TYPE.get_entity_name(), **locals())
    def _map_query(self, _api, lun_container, volume):
        name = lun_container.name if lun_container else None
        lun_containers = self._query(_api, name=name)
        if volume:
            return [Bunch({self.ENTITY_TYPE.get_entity_name() : lun_container.name,
                           "lun" : mapping.lun,
                           "volume" : volume.name})
                    for lun_container in lun_containers
                    for mapping in lun_container.luns
                    if mapping.volume_id == volume.id]
        else:
            return [Bunch({self.ENTITY_TYPE.get_entity_name() : lun_container.name,
                           "lun" : mapping.lun,
                           "volume" : self._get_volume_name(_api, mapping.volume_id)})
                    for lun_container in lun_containers
                    for mapping in lun_container.luns]
    def _get_volume_name(self, _api, volume_id):
        # TODO: Move this to volume.py
        return _api.get("volumes/{}".format(volume_id)).name

@commandgroup(prefix="host", provides=(Host,))
class HostCommands(HostCommandsBase):
    ENTITY_TYPE = Host
    def _host_query_hook(self, hosts):
        return [Bunch(name=host.name,
                      luns=len(host.luns),
                      fc_ports=fc_port_count,
                      created_at=host.created_at,
                      )
                for host in hosts
                for fc_port_count, iscsi_port_count in self._count_ports(host)]
    @command(types=(Name,))
    def create(self, _api, name):
        return self._create(_api, name=name)
    @command(types=(Host, Port),
             substitute=dict(host="name"))
    def add_port(self, _api, host, port):
        type = PortType.TYPES.FC
        _api.post(u"hosts/{}/ports".format(host.id), dict(address=port, type=type))
        return '{} port "{}" added to host "{}"'.format(type.upper(), port, host.name)
    @command(types=(Host, Port),
             substitute=dict(host="name"))
    def remove_port(self, _api, host, port):
        type = PortType.TYPES.FC
        _api.delete(u"hosts/{}/ports/{}/{}".format(host.id, type, port))
        return '{} port "{}" removed from host "{}"'.format(type.upper(), port, host.name)
    @command(types=(Host, Name),
             substitute=dict(host="name"))
    def rename(self, _api, host, new_name):
        return self._rename(_api, host, new_name)
    @command(types=(Host,),
             substitute=dict(host="name"))
    def delete(self, _api, host):
        return self._delete(_api, host)
    @command(types=(Host, Volume, Int),
             substitute=dict(host="name"))
    def map(self, _api, host, volume, lun=None):
        return self._map(_api, host, volume, lun)
    @command(types=(Host, Volume, Int),
             substitute=dict(host="name"))
    def unmap(self, _api, host, volume=None, lun=None):
        return self._unmap(_api, host, volume, lun)
    def _count_ports(self, host):
        fc, iscsi = 0, 0
        for port in host.ports:
            if port.type.lower() == "fc":
                fc += 1
            elif port.type.lower() == "iscsi":
                iscsi += 1
        return ((fc, iscsi),)
    @command(types=(Host,),
             substitute=dict(host="name"))
    @render_collection(entity=Host)
    def query(self, _api, host=None):
        if host is not None:
            return self._host_query_hook(self._query(_api, host))
        return self._paged_query(_api, hook=self._host_query_hook)
    @command(types=(Host, Port),
             substitute=dict(host="name"))
    @render(order=["host"])
    def port_query(self, _api, host=None, port=None):
        name = host.name if host else None
        hosts = self._query(_api, name=name)
        if port:
            port = port.replace(":", "")
            return [Bunch(host=host.name,
                          port=hostport.address,
                          type=hostport.type)
                    for host in hosts
                    for hostport in host.ports if port == hostport.address.replace(":", "")]
        else:
            return [Bunch(host=host.name,
                          port=hostport.address,
                          type=hostport.type)
                    for host in hosts
                    for hostport in host.ports]
    @command(types=(Host, Volume),
             substitute=dict(host="name"))
    @render_collection(alignment="left", order=["host"])
    def map_query(self, _api, host=None, volume=None):
        return self._map_query(_api, host, volume)
