from infi.shell.commands import commandgroup, command, render
from infi.shell.commands.types import Str, Int, Bool
from infi.shell.utils.python import getattr_deep
from ..exceptions import APIFailure
from ..utils import PossiblyNone
from ..system import system
from .types import Pool
from . import superuser
from collections import OrderedDict

class FieldGetter(object):
    def __init__(self, obj, default='N/A'):
        self.obj = obj
        self.default = default
    def __getitem__(self, item):
        return getattr_deep(self.obj, item, self.default)

@commandgroup(prefix="system")
class SystemCommands(object):
    CAPACITY_DEDUCTION = 100000000000;  # See CLI-472, CLI-473
    @command
    def stats_query(self, _api):
        return _api.get("system/stats")
    def _get_capacity_info(self, _api, system):
        total_capacity = system.capacity.total_hard_capacity
        free_capacity = system.capacity.free_hard_capacity
        # Get free capacity of largest pool (proprietary behavior for single-pool Infiniboxes)
        pools = Pool.paged_query(_api)
        if pools.total_objects:
            customer_pool = reduce((lambda p1, p2: p1 if p1.hard_capacity > p2.hard_capacity else p2), pools)
            total_capacity = customer_pool.hard_capacity
            free_capacity = customer_pool.disk_unused
        return max(total_capacity - self.CAPACITY_DEDUCTION, 0), max(free_capacity - self.CAPACITY_DEDUCTION, 0)
    @command
    @render(individually=True)
    def info(self, _api):
        _system = _api.get("system")
        _system.capacity = PossiblyNone.wrap_items(_system.capacity)
        total_capacity, free_capacity = self._get_capacity_info(_api, _system)
        result = OrderedDict((
            ("name", _system.name),
            ("serial", _system.serial_number),
            ("product_id", _system.product_id),
            ("version", _system.version),
            ("revision", _system.revision),
            ("deployment_id", _system.deployment_id),
            ("operational_state", _system.operational_state.state),
            ("total_capacity", total_capacity),
            ("free_capacity", free_capacity),
            ))
        for collection in ("pools", "volumes", "hosts"):
            result[collection] = self._get_total_objects(_api, collection)
        # Extra info for infinidat technicians
        if system.user.role.is_infinidat():
            result["uptime"] = _system.uptime
            for field, value in sorted(PossiblyNone.wrap_items(_system.health_state or {}).iteritems(), key=lambda pair: pair[0]):
                if isinstance(value, (dict, list, set, tuple)):
                    continue
                result[field] = value
        return result
    def _get_total_objects(self, _api, collection):
        try:
            return _api.get("{}/?page_size=1".format(collection)).metadata.number_of_objects
        except APIFailure:
            return "N/A"
    @command
    def log_mark(self, _api):
        _api.post("logs/mark", {})
        return "Logs marked"
    @command(types=(Str, Int))
    def log_freeze(self, _api, cause=None, duration=None):
        _api.post("logs/freeze", dict(cause=cause, duration=duration))
        return "Logs frozen"
    @command(types=(Bool,))
    def shutdown(self, _api, poweroff=False):
        operation = "shutdown_and_poweroff" if poweroff else "shutdown"
        _api.post("system/operational_state/" + operation)
        return "Shutting down system"
    @superuser
    @command
    def emergency_shutdown(self, _api):
        _api.post("system/operational_state/emergency_shutdown")
        return "SYSTEM GOING INTO EMERGENCY SHUTDOWN"
    @superuser
    @command
    def activate(self, _api):
        _api.put("config/mgmt/state.pause_startup_in", "ACTIVE")
        return "System undergoing activation"
