import capacity
import inspect
import re
from bunch import Bunch

from infi.shell.commands.types import Int, Measure, RegexType, EnumType, Dynamic
from infi.shell.utils import decamelize, SpellingCorrector
from infi.shell.exceptions import TooManyPossibleValues

from ..exceptions import BadURI, NoSuchEntity, InvalidConstant
from ..infinishell_config import config

class Name(RegexType):
    ALLOWED_SYMBOLS = "^&'@()[],$=!-#{}%.+~_ "
    def _get_pattern(self):
        return "^[0-9a-zA-Z{}]+$".format(re.escape(self.ALLOWED_SYMBOLS))
    def _get_error(self):
        return "contains illegal characters"
    def get_description(self):
        return "English alpha-numerics, and the following symbols: {}".format(self.ALLOWED_SYMBOLS)
class ProvType(EnumType):
    def _get_values(self):
        return "thin", "thick"
class PortType(EnumType):
    TYPES = Bunch(
        FC="fc",
        ISCSI="iscsi",
        )
    def _get_values(self):
        return self.TYPES.values()
class Port(RegexType):
    def _get_pattern(self):
        patterns = ("([0-9a-fA-F]{2}:){7}[0-9a-fA-F]{2}",
                    "[0-9a-fA-F]{16}")
        return "^({})$".format("|".join(patterns))
    def _get_error(self):
        return "is not a valid port address"
    def get_description(self):
        return "A 16-digit hexadecimal number, optionally divided by colons to groups of two digits"
class Size(Measure):
    _units = dict((unit, cap)
                  for cap, units in {capacity.byte : ("b"),
                                     capacity.KB : ("k", "kb"),
                                     capacity.MB : ("m", "mb"),
                                     capacity.GB : ("g", "gb"),
                                     capacity.TB : ("t", "tb"),
                                     capacity.PB : ("p", "pb")}.iteritems()
                  for unit in units)
    def _verify(self, value):
        error = super(Size, self)._verify(value)
        if error:
            return error
        if isinstance(value, float):
            return self._get_error()
    def _process_actual(self, size, unit):
        if not unit:
            return size if not isinstance(size, float) else None
        unit = unit.lower()
        if unit == "b" and isinstance(size, float):
            return None
        try:
            cap = self._units[unit]
            return long((size * cap).roundup(capacity.byte) / capacity.byte)
        except:
            return None
    def _get_error(self):
        return "is not a valid size"
    def _is_negative_enabled(self):
        return False
    def _get_fraction_digits(self):
        return 2
    def _get_units(self):
        return self._units.iterkeys()
    def get_description(self):
        return "A positive number (up to 2 digits beyond the decimal point),"\
               " optionally followed by a unit; e.g. 3000000m, 2000GB, 1.5T"
class Lun(Int):
    pass
class ObjectID(RegexType):
    def _get_pattern(self):
        return "^[0-9a-fA-F]{8}-?[0-9a-fA-F]{4}-?[0-9a-fA-F]{4}-?[0-9a-fA-F]{4}-?[0-9a-fA-F]{12}$"
    def _get_error(self):
        return "is not a UUID"
    def get_description(self):
        return "A 16-digit hexadecimal number, optionally divided by dashes to groups of 8-4-4-12"
class ISRPolicy(EnumType):
    def _get_values(self):
        return "hourly", "daily", "week_scaled", "year_scaled"
class SystemModule(EnumType):
    def _get_values(self):
        return ('xn', 'sys', 'event', 'config', 'comm', 'port', 'cache', 'emul',
                'destage', 'snap', 'raid', 'dev', 'ctrl', 'mgmt', 'all')
class LogLevel(EnumType):
    def _get_values(self):
        return "debug", "trace", "info", "warn", "error", "fatal"

### SPECIAL TYPES ###

class Partial(object):
    _partials = {}
    @classmethod
    def _create_partial(_cls, _entity_class):
        # TODO: This is a problem. Since it's a different class, it doesn't
        # share its list of value providers with its parent class...
        class PartialEntity(_entity_class):
            _is_partial = True
            def get_possible_values(self, api, name_starts_with=None):
                return super(PartialEntity, self).get_possible_values(api, name_starts_with, _entity_class)
        PartialEntity.__name__ = _entity_class.__name__
        return PartialEntity
    @classmethod
    def _get_partial(cls, entity_class):
        partial = cls._partials.get(entity_class)
        if partial is None:
            partial = cls._partials[entity_class] = cls._create_partial(entity_class)
        return partial
    def __get__(self, instance, owner):
        return self._get_partial(owner)

class Entity(Dynamic):
    CACHEABLE = True
    URI = None
    SEARCHABLE = True
    PAGED = True
    ID_FIELD = "id"
    NAME_FIELD = "name"
    ENTITY_NAME = None
    ENTITY_MGMT_NAME = None
    ENTITY_DISPLAY_NAME = None
    DESCRIPTION = None
    QUERY_PARAMS = None
    PARTIAL = Partial()
    _is_partial = False
    MAX_POSSIBLE_VALUES = 200
    @classmethod
    def get_entity_name(cls):
        if not hasattr(cls, "__entity_name"):
            cls.__entity_name = getattr(cls, "ENTITY_NAME", None) or cls.__name__.lower()
        return cls.__entity_name
    @classmethod
    def get_entity_mgmt_name(cls):
        if not hasattr(cls, "__entity_mgmt_name"):
            cls.__entity_mgmt_name = (getattr(cls, "ENTITY_MGMT_NAME", None) or cls.get_entity_name()).lower()
        return cls.__entity_mgmt_name
    @classmethod
    def get_display_name(cls):
        if not hasattr(cls, "__entity_display_name"):
            cls.__entity_display_name = (getattr(cls, "ENTITY_DISPLAY_NAME", None) or decamelize(cls.__name__)).title()
        return cls.__entity_display_name
    @classmethod
    def get_description(cls):
        if not hasattr(cls, "__description"):
            cls.__description = getattr(cls, "DESCRIPTION", None) or "Name of an existing {}".format(cls.get_display_name().lower())
        return cls.__description
    @classmethod
    def get_uri(cls, *subresource):
        if not hasattr(cls, "__uri"):
            cls.__uri = cls.URI or (cls.__name__.lower() + "s")
        if not subresource:
            return cls.__uri
        return "{}/{}".format(cls.__uri, "/".join(str(s) for s in subresource))
    def _get_dynamic(self, api, arg, value):
        return self._fetch(api, arg, value)
    @classmethod
    def _fetch(cls, api, arg, value):
        if arg.endswith("_id"):
            arg = arg[:-3]
            return arg, cls.get_by_id(api, value)
        return arg, cls.get_by_name(api, value)
    @classmethod
    def _get_query_params(cls):
        return cls.QUERY_PARAMS
    @classmethod
    def _compute_query_params(cls, params):
        base_params = cls._get_query_params()
        if not base_params:
            return params
        if not params:
            return base_params
        params = dict(params)
        params.update(base_params)
        return params
    @classmethod
    def _compute_query_hook(cls, hook):
        if not hook:
            return hook
        def wrapper_hook(results):
            results = cls._query_hook(results)
            results = hook(results)
            return results
        return wrapper_hook
    @classmethod
    def paged_query(cls, api, params=None, hook=None, **kwargs):
        params = cls._compute_query_params(params)
        hook = cls._compute_query_hook(hook)
        return api.paged_query(cls.get_uri(), params=params, hook=hook, **kwargs)
    @classmethod
    def query(cls, api, params=None, hook=None, **kwargs):
        params = cls._compute_query_params(params)
        hook  = cls._compute_query_hook(hook)
        result = api.get(cls.get_uri(), params=params, **kwargs)
        if hook:
            result = hook(result)
        return result
    @classmethod
    def iter_all(cls, api, params=None, page_size=None, hook=None):
        if cls.PAGED:
            return cls.paged_query(api, params=params, page_size=page_size, hook=hook).iter_all_objects()
        else:
            return iter(cls.get_all(api, params=params, hook=hook))
    @classmethod
    def get_all(cls, api, params=None, hook=None):
        if cls.PAGED:
            return list(cls.iter_all(api, params=params, page_size=config.api_max_page_size, hook=hook))
        else:
            return cls.query(api, params=params, hook=hook)
    @classmethod
    def _query_hook(cls, results):
        return results
    @classmethod
    def search(cls, api, query="", page_size=None):
        assert cls.SEARCHABLE, cls.__name__ + " is not searchable"
        return api.paged_query(
            "search",
            page_size=page_size or config.api_default_search_page_size,
            params=dict(
                type=cls.get_entity_mgmt_name(),
                query=query,
                ),
            hook=cls._search_hook,
            )
    @classmethod
    def _search_hook(cls, results):
        return results
    @classmethod
    def get_by_id(cls, api, id):
        try:
            entity = api.get("{}/{}".format(cls.get_uri(), id))
            cls._validate_fetched_entity(entity)
            return entity
        except BadURI:
            raise NoSuchEntity(u'{} with ID {}'.format(cls.get_display_name(), id))
    @classmethod
    def get_by_name(cls, api, name):
        if not name and name.strip():
            raise NoSuchEntity(u"{} name must not be empty".format(cls.get_entity_name()))
        entity = cls._fetch_entity_by_name(api, name)
        if entity is not None:
            cls._validate_fetched_entity(entity)
            return entity
        raise NoSuchEntity(u'{} "{}"'.format(cls.get_display_name().lower(), name))
    @classmethod
    def _fetch_entity_by_name(cls, api, name):
        entities = cls.get_all(api, params=dict(name=name))
        if entities is not None:
            for entity in entities:
                if entity[cls.NAME_FIELD] == name:
                    return entity
        return None
    @classmethod
    def _validate_fetched_entity(cls, entity):
        pass
    def get_possible_values(self, api, name_starts_with=None, type_cls=None):
        type_cls = type_cls or type(self)
        if type_cls.SEARCHABLE:
            return self._get_possible_values_by_search(api, name_starts_with, type_cls)
        return self._get_possible_values_by_query(api, name_starts_with, type_cls)
    def _get_possible_values_by_search(self, api, name_starts_with, type_cls):
        name_starts_with = name_starts_with or ""
        query = api.paged_query(
            "search",
            page_size=config.api_default_search_page_size,
            params=dict(
                type=type_cls.get_entity_mgmt_name(),
                query=name_starts_with,
                ),
            hook=self._search_hook,
            )
        if query.total_objects > self.MAX_POSSIBLE_VALUES:
            raise TooManyPossibleValues
        return [entity_info.name for entity_info in query.iter_all_objects() if entity_info.name.startswith(name_starts_with)]
    def _get_possible_values_by_query(self, api, name_starts_with, type_cls):
        cursor = (entity[type_cls.NAME_FIELD] for entity in type_cls.iter_all(api))
        if name_starts_with:
            cursor = (name for name in cursor if name.startswith(name_starts_with))
        return cursor
class Volume(Entity):
    class __metaclass__(getattr(Entity, '__metaclass__', type)):
        def __new__(mcls, name, bases, attrs):
            cls = getattr(Entity, '__metaclass__', type).__new__(mcls, name, bases, attrs)
            # Map volume type names to volume entity classes
            if name != "Volume" and cls.VOLUME_TYPE:
                Volume._types[cls.VOLUME_TYPE] = cls
            return cls
    URI = "volumes"
    VOLUME_TYPE = None
    ENTITY_MGMT_NAME = "VOLUME"
    DESCRIPTION = None  # Do not override this
    DESCRIPTION_SUBJECT = "volume, snapshot or clone"  # Override this
    _types = {}
    @classmethod
    def _type_name_to_type(cls, type_name):
        entity_type = cls._types.get(type_name.upper())
        if entity_type is None:
            raise AssertionError("Type not registered: " + type_name)
        return entity_type
    @classmethod
    def get_description(cls):
        if not hasattr(cls, "__description"):
            cls.__description = "Name of an existing {}".format(cls.DESCRIPTION_SUBJECT)
        return cls.__description
class VolumeSubtypeMixin(object):
    @classmethod
    def _fetch_entity_by_name(cls, api, name):
        # Call API explicitly instead of get_all() or iter_all(),
        # so that _get_query_params() will not be used
        entities = api.paged_query(cls.get_uri(), params=dict(name=name))
        if entities is not None:
            for entity in entities:
                if entity.name == name:
                    return entity
        return None
class VolumeSubtypeBase(VolumeSubtypeMixin, Volume):
    @classmethod
    def _validate_fetched_entity(cls, entity):
        if entity.type.lower() != cls.VOLUME_TYPE.lower():
            raise NoSuchEntity('"{name}" is a {type}, not a {master}'.format(
                name=entity.name,
                type=cls._type_name_to_type(entity.type).get_display_name().lower(),
                master=cls.get_display_name().lower(),
                ))
    @classmethod
    def _get_query_params(cls):
        return dict(
            type=cls.VOLUME_TYPE.lower(),
            )
    @classmethod
    def _search_hook(cls, entities):
        return [e for e in entities if e.properties.volume_type.lower() == cls.VOLUME_TYPE.lower()]
class MasterVolume(VolumeSubtypeBase):
    VOLUME_TYPE = "MASTER"
    DESCRIPTION_SUBJECT = "master volume"
class Snapshot(VolumeSubtypeBase):
    VOLUME_TYPE = "SNAP"
    DESCRIPTION_SUBJECT = "snapshot"
class Clone(VolumeSubtypeBase):
    VOLUME_TYPE = "CLONE"
    DESCRIPTION_SUBJECT = "clone"
class NonSnapshot(VolumeSubtypeMixin, Volume):
    VOLUME_TYPE = None
    ENTITY_DISPLAY_NAME = "Non-Snapshot"
    DESCRIPTION_SUBJECT = "volume or clone"
    @classmethod
    def _validate_fetched_entity(cls, entity):
        if entity.type.lower() == Snapshot.VOLUME_TYPE.lower():
            raise NoSuchEntity('"{name}" is a {snapshot}, not a {volume} or {clone}'.format(
                name=entity.name,
                snapshot=Snapshot.get_display_name().lower(),
                volume=Volume.get_display_name().lower(),
                clone=Clone.get_display_name().lower(),
                ))
    @classmethod
    def _query_hook(cls, results):
        return [e for e in results if e.type.lower() != Snapshot.type.lower()]
    @classmethod
    def _search_hook(cls, entities):
        return [e for e in entities if e.properties.volume_type.lower() != Snapshot.VOLUME_TYPE.lower()]
class Pool(Entity):
    pass
class Host(Entity):
    pass
class Cluster(Entity):
    pass
class Event(Entity):
    SEARCHABLE = False
class EventRule(Entity):
    URI = "events/mail"
    SEARCHABLE = False
    @classmethod
    def _fetch_entity_by_name(cls, api, name):
        ''' Workaround for INFINIBOX-5507 '''
        for rule in cls.iter_all(api):
            if rule.name == name:
                return rule
        return None

class Node(Entity):
    URI = "components/nodes"
    CACHEABLE = True
    SEARCHABLE = False
    PAGED = False
    def get_description(self):
        return "Node name"

class Enclosure(Entity):
    SEARCHABLE = False
    PAGED = False
    URI = "components/enclosures"

class Drive(Entity):
    SEARCHABLE = False
    PAGED = False
    @classmethod
    def get_all(cls, api):
        enclosures = Enclosure.iter_all(api)
        drives = [cls._handle_drive(e, d) for e in enclosures for d in e.drives]
        return drives
    @classmethod
    def _handle_drive(cls, enclosure, drive):
        drive.index = "E{}D{:02}".format(drive.enclosure_index, drive.drive_index)
        return drive
    @classmethod
    def iter_all(cls, api):
        return iter(cls.get_all(api))
    @classmethod
    def get_by_enclosure_and_drive(cls, api, enclosure_id, drive_id):
        return api.get("{}/{}/drives/{}".format(Enclosure.get_uri(), enclosure_id, drive_id))

class User(Entity):
    SEARCHABLE = False
    ID_FIELD = "uid"
    NAME_FIELD = "uid"
    @classmethod
    def _fetch_entity_by_name(cls, api, name):  # Workaround for INFINIBOX-6242
        entities = cls.get_all(api)
        if entities is not None:
            for entity in entities:
                if entity[cls.NAME_FIELD] == name:
                    return entity
        return None

class ConstantEntity(Entity):
    _subclasses = []
    class __metaclass__(getattr(Entity, '__metaclass__', type)):
        def __new__(mcls, name, bases, attrs):
            new_class = getattr(Entity, '__metaclass__', type).__new__(mcls, name, bases, attrs)
            if name != "ConstantEntity":
                ConstantEntity._subclasses.append(new_class)
            return new_class
    CACHEABLE = True
    FETCH_ONCE = True
    SEARCHABLE = False
    PAGED = False
    URI = None # This must be overridden
    ATTR = None # This may be overridden
    _values = None
    _spelling_corrector = None
    @classmethod
    def _request_values(cls, api):
        values = api.get(cls.URI)
        if cls.ATTR:
            for attr in cls.ATTR.split('.'):
                values = getattr(values, attr)
        return values
    @classmethod
    def _ensure_values(cls, api):
        if cls._values is None:
            values = cls._request_values(api)
            values.sort()
            cls._values = values
            cls._spelling_corrector = SpellingCorrector(cls._values)
    @classmethod
    def invalidate_cache(cls):
        """ Invalidate cache of constant values, allowing them to be re-fetched
        from server. """
        cls._values = None
    @classmethod
    def invalidate_all_caches(cls):
        """ Invalidate cache of constant values for all constants, allowing them
        to be re-fetched from server. This method is present only in
        ConstantEntity, and absent from all its subclasses. """
        for subclass in cls._subclasses:
            subclass.invalidate_cache()
    @classmethod
    def _fetch(cls, api, arg, value):
        cls._ensure_values(api)
        for const_value in cls._values:
            if const_value.lower() == value.lower():
                return arg, const_value
        suggestions = cls._spelling_corrector.correct(value)
        msg = u"{} is not a valid {}".format(value, cls.get_display_name())
        if suggestions:
            msg += u". Did you mean: {}?".format(u", ".join(suggestions))
        raise InvalidConstant(msg)
    @classmethod
    def iter_all(cls, api):
        cls._ensure_values(api)
        return iter(cls._values)
    @classmethod
    def get_all(cls, api):
        return list(cls.iter_all(api))
    def get_possible_values(self, api, name_starts_with=None, type_cls=None):
        it = self.iter_all(api)
        if name_starts_with:
            it = (value for value in it if value.startswith(name_starts_with))
        return it
class EventLevel(ConstantEntity):
    URI = "events/types"
    ATTR = "levels"
    def get_description(self):
        return "Event severity level"
class EventCode(ConstantEntity):
    URI = "events/types"
    ATTR = "codes"
    def get_description(self):
        return "Event type name"
class EventVisibility(ConstantEntity):
    URI = "events/types"
    ATTR = "visibilities"
    def get_description(self):
        return "Event visibility scope"
class EventReporter(ConstantEntity):
    URI = "events/types"
    ATTR = "reporters"
    def get_description(self):
        return "Event reporter"
class ConfigParameter(ConstantEntity):
    @classmethod
    def get_name(cls):
        return "PARAM"
    def get_description(self):
        return "Name of a configuration parameter"
    @classmethod
    def _request_values(cls, api):
        categories = api.get("config")
        names = []
        for category_name, category in categories.iteritems():
            names.extend(cls._get_param_names({ category_name : category }))
        names.sort()
        return names
    @classmethod
    def _get_param_names(cls, category):
        for name, value in category.iteritems():
            if isinstance(value, dict):
                for nested_name in cls._get_param_names(value):
                    yield "{}/{}".format(name, nested_name)
            else:
                yield name

class ObjType(EnumType):
    TYPES = {c.__name__.lower() : c for c in globals().values()
             if inspect.isclass(c)
             and issubclass(c, Entity)
             and c is not Entity}
    TYPE_NAMES = TYPES.keys()
    @classmethod
    def get_type(cls, name):
        return cls.TYPES[name.lower()]
    def _get_values(self):
        return self.TYPE_NAMES

