from infi.shell.commands import commandgroup, command, render
from infi.shell.commands.types import EmailAddress
from infi.shell.rlcontext import get_raw_input_functions
from infi.shell.exceptions import CommandFailed, CommandAborted
from infi.shell.shell import get_active_shell

from ..system import system
from .types import User
from . import EntityCommands

@commandgroup(prefix="user")
class UserCommands(EntityCommands):
    ENTITY_TYPE = User

    @command(
        types=(User,),
        substitute=dict(user="name"),
        description="Display all users, or a single user"
        )
    @render(
        order="uid role email".split(),
        substitute=dict(uid="name"),
        )
    def query(self, _api, user=None):
        result = self._query(_api, user)
        result = sorted(result, key=lambda user: user.uid)
        return result

    def _input_new_password(self, username=None):
        getpass = get_raw_input_functions()[1]
        if username is not None and not system.check_credentials(username, getpass("Current password: ")):
            if username == system.user.username:
                get_active_shell().error("This is not your current password, terminating shell")
                raise EOFError
            raise CommandFailed("This is not the current password, please try again")
        password = getpass(prompt="New password: ")
        if not password.strip():
            raise CommandAborted("Password is empty, aborting")
        if password != getpass(prompt="Re-enter new password: "):
            raise CommandFailed("Passwords differ from one another, please try again")
        return password

    @command(
        types=(User, EmailAddress),
        substitute=dict(user="name"),
        description="Update existing user",
        )
    def update(self, _api, user, email=None):
        _api.put(User.get_uri(user.uid), dict(
            email=email,
            ))
        return self._result_message(user, "updated")

    @command(
        types=(User,),
        substitute=dict(user="name"),
        description="Interactively reset password of existing user",
        )
    def change_password(self, _api, user=None):
        user = user or system.user.attributes
        password = self._input_new_password(user.uid)
        _api.put(User.get_uri(user.uid), dict(password=password))
        if system.user.username == user.uid:
            system.login(user.uid, password)
        return self._result_message(user, "updated")

    @command(
        types=(User,),
        substitute=dict(user="name"),
        description="Send password reset email to existing user",
        )
    def reset_password(slf, _api, user):
        _api.post(User.get_uri(user.uid, "reset_password"))
        return "Email sent to user"