from bunch import Bunch
from functools import partial
from collections import defaultdict

from infi.shell.exceptions import BadArguments
from infi.shell.utils.python import Lazy
from infi.shell.commands import commandgroup, command
from infi.shell.commands.types import Int

from ..exceptions import NoSuchEntity
from .types import Pool, Volume, Snapshot, Clone, NonSnapshot, MasterVolume, Cluster, Host, Name, Size
from . import EntityCommands, render_collection, superuser, admin_only

EXCLUDED_QUERY_FIELDS = "pool_id parent_id depth has_children data provtype serial num_blocks".split()
QUERY_FIELD_ORDER = "name type size pool write_protected mapped".split()
QUERY_SUBSTITUTES = dict(write_protected="protected")

class VolumeCommandsBase(EntityCommands):
    def _volume_query(self, _api, volume, _with_pools=True, **params):
        hook = None
        if _with_pools:
            lazy_pool_names_by_id = Lazy(lambda: {p.id : p.name for p in Pool.search(_api).iter_all_objects()})
            hook = partial(self._query_hook, lazy_pool_names_by_id)
        return self._query(_api, volume, _hook=hook, **params)
    def _query_hook(self, lazy_pool_names_by_id, volumes):
        for v in volumes:
            v.pool = lazy_pool_names_by_id.value.get(v.pool_id) or "UNKNOWN (ID: {})".format(v.pool_id)
        return volumes
    class VolumeTreeNode(object):
        def __init__(self, volume=None, children=None):
            self.volume = volume
            self.children = [] if children is None else children
            self.depth = 0
            self.parent = None
        def add_child(self, treenode):
            self.children.append(treenode)
            treenode.parent = self
            treenode.depth = self.depth + 1
    def _get_all_volume_trees(self, _api):
        treenodes = defaultdict(self.VolumeTreeNode)
        roots = []
        for v in Volume.iter_all(_api):
            treenode = treenodes[v.id]
            treenode.volume = v
            if not v.parent_id:
                roots.append(treenode)
            else:
                treenodes[v.parent_id].add_child(treenode)
        return roots
    def _get_single_volume_tree(self, _api, volume):
        root = self.VolumeTreeNode(volume)
        self._fill_single_tree(_api, root)
        return [root]
    def _fill_single_tree(self, _api, parentnode):
        for volume in Volume.get_all(_api, params=dict(parent_id=parentnode.volume.id)):
            childnode = self.VolumeTreeNode(volume)
            parentnode.add_child(childnode)
            self._fill_single_tree(_api, childnode)
    def _flatten_tree_for_rendering(self, treenodes, result=None):
        if result is None:
            result = []
        treenodes.sort(key=lambda treenode: treenode.volume.id)
        for treenode in treenodes:
            volume = Bunch((field_name, treenode.volume[field_name]) for field_name in ('type', 'size'))
            volume.name = ('|   ' * treenode.depth) + treenode.volume.name
            result.append(volume)
            self._flatten_tree_for_rendering(treenode.children, result=result)
        return result

def create_common_volume_commands_mixin(VolumeType):

    class CommonVolumeCommandsMixin(object):
        @command(types=(VolumeType, Name),
                 substitute=dict(volume="name"))
        def rename(self, _api, volume, new_name):
            return self._rename(_api, volume, new_name)
        @command(types=(VolumeType,),
                 substitute=dict(volume="name"))
        def delete(self, _api, volume):
            return self._delete(_api, volume)
        def _get_mappable(self, host, cluster):
            if host:
                return Host, host
            if cluster:
                return Cluster, cluster
            raise AssertionError("Undefined mappable type")
        @command(types=(VolumeType, Host, Cluster, Int),
                 substitute=dict(volume="name"))
        def map(self, _api, volume, host=None, cluster=None, lun=None):
            if not host and not cluster:
                raise BadArguments("Either host or cluster must be specified")
            props = {"volume_id" : volume.id}
            if lun is not None:
                props["lun"] = lun
            MappableType, mappable = self._get_mappable(host, cluster)
            result = _api.post("{}/{}/luns".format(MappableType.get_uri(), mappable.id), props)
            lun_text = "LUN {} in ".format(result.lun) if "lun" in result else ""
            return 'Volume "{volume.name}" mapped to {lun_text}{mappable_type} "{mappable.name}"'.format(
                volume=volume,
                lun_text=lun_text,
                mappable_type=MappableType.get_display_name().lower(),
                mappable=mappable,
                )
        @command(types=(VolumeType, Host, Cluster),
                 substitute=dict(volume="name"))
        def unmap(self, _api, volume, host=None, cluster=None):
            if not host and not cluster:
                raise BadArguments("Either host or cluster must be specified")
            MappableType, mappable = self._get_mappable(host, cluster)
            result = _api.delete("{}/{}/luns/volume_id/{}".format(MappableType.get_uri(), mappable.id, volume.id))
            return 'Volume "{volume.name}" unmapped from LUN {lun} in {mappable_type} "{mappable.name}"'.format(
                volume=volume,
                lun=result.lun,
                mappable_type=MappableType.get_display_name().lower(),
                mappable=mappable,
                )
        @command(types=(VolumeType,),
                 substitute=dict(volume="name"))
        def tree(self, _api, volume=None):
            if volume is None:
                roots = self._get_all_volume_trees(_api)
            else:
                roots = self._get_single_volume_tree(_api, volume)
            return self._flatten_tree_for_rendering(roots)

    return CommonVolumeCommandsMixin

def create_writable_volume_commands_mixin(VolumeType):

    class WritableVolumeCommandsMixin(object):
        @command(types=(VolumeType, Volume),
                 substitute=dict(volume="name"))
        def restore(self, _api, volume, snapshot):
            self._update(_api, volume, data=snapshot.data)
            return self._result_message(volume, 'restored to snapshot "{}"'.format(snapshot.name))
        @command(types=(VolumeType,),
                 substitute=dict(volume="name"))
        def write_protect(self, _api, volume):
            return self._update(_api, volume, write_protected=True)
        @command(types=(VolumeType,),
                 substitute=dict(volume="name"))
        def write_enable(self, _api, volume):
            return self._update(_api, volume, write_protected=False)

    return WritableVolumeCommandsMixin

def create_child_volume_commands_mixin(VolumeType, ParentType, parent_type_name):

    class ChildVolumeCommandsMixin(object):
        @command(types=(VolumeType, ParentType, Pool),
                 substitute=dict(volume="name", parent=parent_type_name))
        @render_collection(entity=Volume,
                           excluded=EXCLUDED_QUERY_FIELDS,
                           order=QUERY_FIELD_ORDER,
                           substitute=QUERY_SUBSTITUTES)
        def query(self, _api, volume=None, parent=None, pool=None):
            return self._volume_query(_api, volume, parent_id=parent, pool_id=pool, type=VolumeType.VOLUME_TYPE)
        @admin_only
        @command(name="query",
                 types=(VolumeType, ParentType),
                 substitute=dict(volume="name", parent=parent_type_name))
        @render_collection(entity=Volume,
                           excluded=EXCLUDED_QUERY_FIELDS + ["pool", "used_size_on_disk"],
                           order=QUERY_FIELD_ORDER,
                           substitute=QUERY_SUBSTITUTES)
        def admin__query(self, _api, volume=None, parent=None):
            return self._volume_query(_api, volume, _with_pools=False, parent_id=parent, type=VolumeType.VOLUME_TYPE)
        @command(types=(ParentType, Name),
                 substitute=dict(volume=parent_type_name))
        def create(self, _api, volume, name):
            parent = self._create_and_return_entity(_api, parent_id=volume.id, name=name)
            return self._result_message(parent, "created", subject=VolumeType.get_display_name().lower())

    return ChildVolumeCommandsMixin

@commandgroup(prefix="volume")
class VolumeCommands(VolumeCommandsBase,
                     create_common_volume_commands_mixin(Volume),
                     create_writable_volume_commands_mixin(Volume)):
    ENTITY_TYPE = Volume
    @superuser
    @command(types=(Volume, Pool),
             substitute=dict(volume="name"))
    @render_collection(entity=Volume,
                       excluded=EXCLUDED_QUERY_FIELDS,
                       order=QUERY_FIELD_ORDER,
                       substitute=QUERY_SUBSTITUTES)
    def query(self, _api, volume=None, pool=None):
        return self._volume_query(_api, volume, pool_id=pool)
    @admin_only
    @command(name="query",
             types=(Volume,),
             substitute=dict(volume="name"))
    @render_collection(entity=Volume,
                       excluded=EXCLUDED_QUERY_FIELDS + ["pool", "used_size_on_disk"],
                       order=QUERY_FIELD_ORDER,
                       substitute=QUERY_SUBSTITUTES)
    def admin__query(self, _api, volume=None):
        return self._volume_query(_api, volume, _with_pools=False)
    @superuser
    @command(types=(Name, Size, Pool))
    def create(self, _api, name, size, pool):
        return self._create(_api, name=name, size=size, pool_id=pool.id, provtype='THICK')
    @admin_only
    @command(name="create",
             types=(Name, Size))
    def admin__create(self, _api, name, size):
        try:
            pool = next(iter(Pool.iter_all(_api)))
        except StopIteration:
            raise NoSuchEntity("Default pool does not exist")
        return self.create(_api, name=name, size=size, pool=pool)
    @superuser
    @command(types=(Volume, Pool),
             substitute=dict(volume="name"))
    def change_pool(self, _api, volume, new_pool):
        self._update(_api, volume, pool_id=new_pool.id)
        return self._result_message(volume, 'moved to pool "{}"'.format(new_pool.name))
    @command(types=(MasterVolume, Size),
             substitute=dict(volume="name"))
    def resize(self, _api, volume, new_size):
        return self._update(_api, volume, size=new_size)

@commandgroup(prefix="snapshot")
class SnapshotCommands(VolumeCommandsBase,
                       create_common_volume_commands_mixin(Snapshot),
                       create_child_volume_commands_mixin(Snapshot, NonSnapshot, "volume")):
    ENTITY_TYPE = Snapshot

@commandgroup(prefix="clone")
class CloneCommands(VolumeCommandsBase,
                    create_common_volume_commands_mixin(Clone),
                    create_writable_volume_commands_mixin(Clone),
                    create_child_volume_commands_mixin(Clone, Snapshot, "snapshot")):
    ENTITY_TYPE = Clone
