from __future__ import print_function

import infi.shell.utils.colorama_patches  # noqa

from . import debug
from . import raw_mode
from infi.shell.commands.commands import Commands
from infi.shell.parser import Parser, ShellParser, CHARS as PARSER_CHARS
from infi.shell.config import config, default_config, write_default_config, initialize_config
from infi.shell.shell import Executor
from infi.shell.utils.filesystem import datadir
from infi.shell.exceptions import FilesystemException
from infi.shell.formatters import format_exception
from .shell import InfiniboxShell
from . import VERSION
from .api import JsonRestAPI
from .args_processor import InfiniboxArgsProcessor
from .environment import variables as env
from .commands.pool import PoolCommands
from .commands.system import SystemCommands
from .commands.node import NodeCommands
from .commands.drive import DriveCommands
from .commands.enclosure import EnclosureCommands
from .commands.volume import VolumeCommands, SnapshotCommands, CloneCommands
from .commands.host import HostCommands
from .commands.metadata import MetadataCommands
from .commands.events import EventCommands, EventRuleCommands
from .commands.magic import InfiniboxMagicCommands
from .commands.general import GeneralCommands
from .commands.cluster import ClusterCommands
from .commands.config import ConfigCommands
from .commands.user import UserCommands
from .exceptions import (
    InvalidOptions,
    ConnectionFailed,
    ServiceTemporarilyUnavailable,
    InternalServerError,
    ManagementException,
    )
from .formatters import InfinidatExceptionFormatter
from .renderer import InfiniboxRenderer, InfiniboxShellRenderer, CSVRenderer
from . import infinishell_config # Import explicitly to apply Infinibox configuration # noqa
from .system.system import System, push_system
from optparse import OptionParser, IndentedHelpFormatter
import sys
import traceback
import logging
from collections import namedtuple
import time
import socket
import itertools
import textwrap
from getpass import getpass

env.set_prefix("INFINISHELL")

PRODUCT_NAME = "InfiniBox"
WELCOME_MSG = "Infinidat Ltd. Command Line Utility v{}".format(VERSION)
WELCOME_BANNER = "\n{}\n".format(WELCOME_MSG)
EXIT_MSG = "Goodbye!"

BASE_URI = "api"
REST_BASE_URI = "rest"
INTERNAL_BASE_URI = "internal"

DATA_DIR = "~/.infinidat/infinishell"
CONFIG_FILE = "infinishellrc"

ENCODING = "utf-8"
COLOR_OPTIONS = ["always", "never", "auto"]
NO_VALUE = "__no_value__"

COMMANDS = (
    InfiniboxMagicCommands(),
    SystemCommands(),
    NodeCommands(),
    DriveCommands(),
    EnclosureCommands(),
    VolumeCommands(),
    SnapshotCommands(),
    CloneCommands(),
    PoolCommands(),
    HostCommands(),
    ClusterCommands(),
    MetadataCommands(),
    EventCommands(),
    EventRuleCommands(),
    ConfigCommands(),
    GeneralCommands(),
    UserCommands(),
    )

FlagOption = namedtuple("FlagOption", ("name", "enabler_option", "disabler_option"))

def output(msg, *args, **kwargs):
    print(msg.format(*args, **kwargs), file=sys.stdout)

def error(msg, *args, **kwargs):
    prefix = "ERROR: "
    prefixes = itertools.chain([prefix], itertools.cycle([" "*len(prefix)]))
    for line in textwrap.wrap(msg.format(*args, **kwargs), 70-len(prefix)):
        print(next(prefixes) + line, file=sys.stderr)

class InfiniShellOptionParser(OptionParser):
    def __init__(self, *args, **kwargs):
        OptionParser.__init__(self, *args, **kwargs)
        self._flag_options = []
    def error(self, msg):
        self.print_usage(sys.stderr)
        raise InvalidOptions(msg.capitalize())
    def add_flag_option(self, flag, enabler_options, enabler_help, disabler_options, disabler_help):
        default_value = getattr(default_config, flag)
        enabler_dest, disabler_dest = self._get_dest_name(enabler_options), self._get_dest_name(disabler_options)
        for options, help, dest, xor in ((enabler_options, enabler_help, enabler_dest, False),
                                         (disabler_options, disabler_help, disabler_dest, True)):
            is_enabled = default_value^xor
            self.add_option(*options, dest=dest, action="store_true", default=False,
                            help="{} (this is the default)".format(help) if is_enabled else help)
        self._flag_options.append(FlagOption(flag, enabler_dest, disabler_dest))
    def add_config_option(self, *args, **kwargs):
        assert args, "Option names are mandatory"
        assert "default" not in kwargs, "Defaults are not allowed in config options"
        key = kwargs.get("dest")
        key = key or kwargs.setdefault("dest", args[-1][2:] if args[-1].startswith("--") else args[-1][1:])
        if "help" in kwargs:
            kwargs["help"] = kwargs["help"].replace("%default", "{}").format(default_config[key])
        kwargs.setdefault("type", "string")
        if kwargs["type"] == "string":
            kwargs["default"] = NO_VALUE
        kwargs.update(
                action="callback",
                callback=self._config_option_callback,
                )
        self.add_option(*args, **kwargs)
    def _config_option_callback(self, option, option_str, value, parser):
        assert option.dest
        config[option.dest] = value
        setattr(parser.values, option.dest, value)
    def _get_dest_name(self, options):
        returned = None
        for option in options:
            if option[0:2] == "--":
                returned = option[2:]
                break
        return (returned or options[0][1:]).replace("-", "_")
    def set_options_in_config(self, values):
        for flag in self._flag_options:
            is_enabled = config[flag.name]
            if (is_enabled and getattr(values, flag.disabler_option) or
                not is_enabled and getattr(values, flag.enabler_option)):
                config[flag.name] = not is_enabled

def display_version_and_exit(*_, **__):
    output(WELCOME_MSG)
    sys.exit(0)

def write_default_config_and_exit(*_, **__):
    if datadir.exists(CONFIG_FILE):
        error("Configuration file already exists. Delete it first.")
    else:
        try:
            config_file_path = datadir.get_path(CONFIG_FILE)
            write_default_config(config_file_path)
            output('Default configuration file generated at "{}"'.format(config_file_path))
            sys.exit(0)
        except OSError as e:
            error("Failed to write configuration file: {}", e)
    sys.exit(2)

def list_env_vars_and_exit(*_, **__):
    def var_to_str(v):
        returned = "{}_{}".format(env.get_prefix(), v.name)
        if v.is_flag:
            returned += " (flag)"
        if v.help:
            returned += ": " + v.help
        if v.default is not None:
            returned += " (default: {})".format(v.default)
        return returned
    env_vars = "\n  ".join(var_to_str(v) for v in env.get_all() if not v.is_internal)
    output('Environment variables:\n  ' + env_vars)
    output('\nPossible flag values: ' + ', '.join(itertools.chain(env.Variable.POSITIVES,
                                                                env.Variable.NEGATIVES)))
    sys.exit(0)

def assert_options(options, args):
    if len(args) > 1:
        raise InvalidOptions("Too many arguments: {}".format(", ".join(repr(arg) for arg in args)), print_usage=True)
    if (options.command, options.file, options.command_help).count(None) < 2:
        raise InvalidOptions("Please specify at most one of the following arguments: -c, -f, -H")
    if options.color not in ([NO_VALUE] + COLOR_OPTIONS):
        raise InvalidOptions("Options for --color are: {}".format(", ".join(COLOR_OPTIONS)))

def handle_debug(options):
    if options.debug:
        debug.on()

def handle_raw_mode(options):
    if options.raw:
        raw_mode.toggle()

def get_script(options):
    if options.command:
        script = options.command
        if PARSER_CHARS.cmd_separator in script:
            raise InvalidOptions("Multiple commands are only allowed in scripts")
    elif options.file:
        with open(options.file) as f:
            script = f.read()
    elif options.command_help:
        script = "help on='{}'".format(options.command_help)
    elif not sys.stdin.isatty():
        script = sys.stdin.read()
    else:
        return None
    return script.decode(ENCODING)

def get_address(args):
    address = args[0] if args else env.get("HOSTNAME", config.default_address)
    if not address:
        raise InvalidOptions("Please specify system address", print_usage=True)
    if ':' in address:
        raise InvalidOptions("Invalid system address: specifying port is not allowed")
    return address

def configure_logging():
# This is disabled because currently there's no need for this kind of logging.
# If anybody ever enables this, don't use config.log_file in the condition,
# because it's already used for activating the command logger.
#    if config.log_file:
#        logging.basicConfig(filename=config.log_file, filemode="a", level=logging.DEBUG,
#                            format="%(asctime)s [%(levelname)s] %(message)s",
#                            datefmt="%Y-%m-%d %H:%M:%S")
#    else:
        logging.disable(logging.FATAL)

def simplify_address(address):
    address = address.lower()
    simplified_address = "".join(c for c in address if c in ("abcdefghijklmnopqrstuvwxyz1234567890-_"))
    return "{}_{:x}".format(simplified_address, abs(hash(address)))

def get_history_filename(address):
    if not config.history:
        return None
    if config.history_file:
        return config.history_file
    simplified_address = simplify_address(address)
    return datadir.get_path("history", simplified_address)

def new_client_id():
    localhost = socket.gethostname()
    return "InfiniShell-{}-{:x}".format(localhost, int(time.time()*1000))

def check_connection(system, address):
    output("Connecting to {}...", address)
    system.api.get("system")
    system.user

def set_username_password(options):
    if env.is_flagged("NO_AUTH"):
        return
    if options.password is not NO_VALUE and options.username is NO_VALUE:
        raise InvalidOptions("Please specify either a username, a username and password, or neither.")
    def set_password():
        config.password = getpass("{} password: ".format(PRODUCT_NAME))
    if not config.username:
        auth = env.get("AUTH")
        if auth:
            auth = auth.split(":", 1)
            if len(auth) != 2:
                raise InvalidOptions('INFINISHELL_AUTH environment variable should contain '
                                     'colon-separated username and password, e.g: "user:pass"')
            config.username, config.password = auth
        else:
            if not sys.stdin.isatty():
                raise InvalidOptions("Credentials not set, and cannot get them while running "
                                     "non-interactively. Use the configuration file (see help), "
                                     "or INFINISHELL_AUTH environment variable.")
            config.username = raw_input("{} username: ".format(PRODUCT_NAME))
            set_password()
    elif not config.password:
        if not sys.stdin.isatty():
            raise InvalidOptions("Only username was set in configuration file, and "
                                 "cannot get password while running non-interactively.")
        set_password()

def get_default_headers():
    headers = dict()
    for header, var in (
            ("X-Forwarded-For", "X_FORWARDED_FOR"),
            ):
        if env.is_set(var):
            headers[header] = env.get(var)
    return headers

def run(address, script=None):
    try:
        try:
            is_interactive = script is None
            ssl_enabled = config.api_ssl and not env.is_flagged("NO_SSL")
            base_api = JsonRestAPI(
                    "{protocol}://{address}:{port}/{base_uri}".format(
                        protocol="https" if ssl_enabled else "http",
                        address=address,
                        port=config.api_ssl_port if ssl_enabled else config.api_port,
                        base_uri=BASE_URI,
                        ),
                    new_client_id(),
                    headers=get_default_headers(),
                    )
            api = base_api.get_sub_resource(REST_BASE_URI, internal=base_api.get_sub_resource(INTERNAL_BASE_URI))

            system = System(api)
            push_system(system)


            if is_interactive:
                output("Connecting to {}...", address)

            # The no-auth mode comes to work around having to authenticate when
            # using InfiniShell from within the Infinibox command line shell
            no_auth = env.is_flagged("NO_AUTH")
            if no_auth:
                system._set_user(uid="infinidat", email="no-reply@infinidat.com", role="Infinidat")
            else:
                system.login(config.username, config.password)  # Also sets API basic auth

            commands = Commands(
                    *COMMANDS,
                    args_processor=InfiniboxArgsProcessor(),
                    default_renderer=InfiniboxRenderer() if script else InfiniboxShellRenderer(),
                    api=api
                )

            parser_kwargs = dict(commands=commands, encoding="utf-8")
            _color_str = config.color.lower()
            _color = True if _color_str == "always" else False if _color_str == "never" else None
            out_stream, err_stream = sys.stdout, sys.stderr
            common_kwargs = dict(
                stream=out_stream,
                paging=config.paging,
                exc_formatter=InfinidatExceptionFormatter(),
                color=_color,
                errors_color=_color,
                )

            def prepare(shell):
                if config.log_file:
                    shell.start_logging(config.log_file)
                if config.csv_output:
                    shell.set_renderer_override(CSVRenderer())

            if is_interactive:
                shell = InfiniboxShell(parser=ShellParser(**parser_kwargs),
                              history_filename=get_history_filename(address),
                              encoding=ENCODING,
                              prompt=address,
                              welcome_msg=WELCOME_BANNER,
                              exit_msg=EXIT_MSG,
                              error_stream=out_stream,  # In interactive mode stderr is not used
                              **common_kwargs
                              )
                prepare(shell)
                shell.start()
                return 0
            else:
                executor = Executor(parser=Parser(**parser_kwargs),
                                    keep_going=config.keep_going,
                                    quiet=config.quiet,
                                    **common_kwargs
                                    )
                prepare(executor)
                success = executor.execute(script)
                return 0 if success else 2
        except:
            # In debug mode, stack trace should be printed for all exceptions
            if env.is_flagged("DEBUG"):
                traceback.print_exc()
            raise
    except (ConnectionFailed, ManagementException) as e:
        error("Unable to reach {}: {}".format(PRODUCT_NAME, str(e)))
        return 2
    except ServiceTemporarilyUnavailable:
        error("{} temporarily unavailable".format(PRODUCT_NAME))
        return 2
    except InternalServerError:
        error("An unknown error occurred on the server side: {}".format(format_exception()))
        return 2
    except FilesystemException as e:
        error(str(e))
        return 1
    except (InvalidOptions, KeyboardInterrupt):
        raise
    except:
        # Save exc_info now, in case other exceptions arise before post mortem
        exc_info = sys.exc_info()
        # For unknown exceptions, stack trace should be printed
        if not env.is_flagged("DEBUG"):
            traceback.print_exc()
        # In debug mode, perform post mortem
        if env.is_flagged("DEBUG"):
            try:
                import ipdb as pdb
            except ImportError:
                import pdb # noqa
            output("\n### POST MORTEM ###")
            pdb.post_mortem(exc_info[2])
        return 1

def _main():
    datadir.init(DATA_DIR)
    config_file_path = datadir.get_path(CONFIG_FILE)
    initialize_config(config_file_path)
    op = InfiniShellOptionParser(prog="infinishell", usage="Usage: %prog [options] system",
                                 epilog='Please note that defaults may be overridden in the configuration '
                                        'file: "{}". Also, some environment variables are available for '
                                        'your convenience. To list them, use the --list-environment option.'
                                        .format(config_file_path),
                                 formatter=IndentedHelpFormatter(max_help_position=30))
    op.add_option("-v", "--version", action="callback", callback=display_version_and_exit,
                  help="Print version and exit")
    op.add_option("-c", "--cmd", dest="command", metavar="CMD", default=None,
                  help="Commands to execute (separated by semicolons)")
    op.add_option("-f", "--file", metavar="FILE", default=None,
                  help="Script file to execute")
    op.add_option("-d", "--debug", action="store_true", default=False,
                  help="Enable debug prints")
    op.add_option("-r", "--raw", action="store_true", default=False,
                  help="Display raw API responses instead of normal rendering")
    op.add_config_option("-u", "--user", dest="username",
                         help="Login username")
    op.add_config_option("-p", "--password",
                         help="Login password")
    op.add_option("-H", dest="command_help", metavar="CMD", default=None,
                  help="Display syntax of the specified command")
    op.add_config_option("-l", "--log", dest="log_file", metavar="FILE",
                         help="Log commands to the specified file (default: %default)")
    op.add_option("--write-default-config", action="callback", callback=write_default_config_and_exit,
                  help='Generate the default configuration file at "{}"'.format(config_file_path))
    op.add_option("--list-environment", action="callback", callback=list_env_vars_and_exit,
                  help="List environment variables")
    op.add_config_option("--color", metavar="WHEN",
                         help="Whether or not to output with color. Options are: always, never, auto (default: %default)")
    op.add_flag_option(flag="keep_going",
                       enabler_options=("-k", "--keep-going"),
                       enabler_help="Keep going when administration errors occur (syntax errors always abort the entire script)",
                       disabler_options=("--fail-fast",),
                       disabler_help="Stop running script when the first administration error occurs")
    op.add_flag_option(flag="quiet",
                       enabler_options=("-q", "--quiet"),
                       enabler_help="No output apart from errors",
                       disabler_options=("--not-quiet",),
                       disabler_help="Normal output, the opposite of --quiet")
    op.add_flag_option(flag="paging",
                       enabler_options=("--with-paging",),
                       enabler_help="Enable paging of large outputs",
                       disabler_options=("--no-paging",),
                       disabler_help="Disable paging of large outputs")
    op.add_flag_option(flag="auto_confirm_prompts",
                       enabler_options=("-y", "--yes"),
                       enabler_help="Automatically confirm prompts (of potentially destructive commands)",
                       disabler_options=("--no",),
                       disabler_help="Automatically reject prompts (of potentially destructive commands)")
    op.add_config_option("--history-file", metavar="PATH", dest="history_file",
                         help="Path of history file, for cross-session history of commands. By default this "
                              "is saved in InfiniShell's data directory, under a unique name generated off of "
                              "the target system address")
    op.add_flag_option(flag="history",
                       enabler_options=("--with-history",),
                       enabler_help="Enable cross-session history of commands",
                       disabler_options=("--no-history",),
                       disabler_help="Disable cross-session history of commands")
    op.add_flag_option(flag="csv_output",
                       enabler_options=("--csv",),
                       enabler_help="CSV output mode (optimize output for easy parsing, see RFC 4180)",
                       disabler_options=("--no-csv",),
                       disabler_help="Disable CSV output mode")
    op.add_config_option("--request-timeout", dest="api_request_timeout_seconds", metavar="SECONDS",
                         type="int",
                         help="Maximal amount of time to wait for management response (default: %default)")
    op.add_flag_option(flag="api_ssl",
                       enabler_options=("--with-ssl",),
                       enabler_help="Encrypt connection using SSL or TLS",
                       disabler_options=("--no-ssl",),
                       disabler_help="Do not encrypt connection")
    try:
        options, args = op.parse_args()
        assert_options(options, args)
        handle_debug(options)
        handle_raw_mode(options)
        op.set_options_in_config(options)
        address = get_address(args)
        script = get_script(options)
        set_username_password(options)
        configure_logging()
        return run(address, script)
    except KeyboardInterrupt:
        error("Aborted")
        return 1
    except Exception as e:
        if isinstance(e, InvalidOptions):
            if e.print_help:
                op.print_help()
            elif e.print_usage:
                op.print_usage()
        if env.is_flagged("DEBUG"):
            traceback.print_exc()
        else:
            error(str(e))
        return 1

def main():
    if env.is_flagged("PROFILER"):
        import cProfile
        cProfile.run("from infinishell.infinibox.main import _main; _main()", sort="cumulative")
        return 0
    else:
        return _main()

if __name__ == '__main__':
    sys.exit(main())

