from infi.shell.console import get_console_size
from infi.shell.renderer import Renderer
from infi.shell.utils import represent
from ..api import PagedQuery
from .. import raw_mode
from ..commands import types
from ..exceptions import UnexpectedAPIResponse
from .values import ValueRenderer
from collections import OrderedDict
from pprint import pformat
from numbers import Number
from itertools import izip, chain
from colorama import Style
import re

__all__ = ['InfiniboxRenderer', 'InfiniboxShellRenderer', 'CollectionRenderer']

"""
InfiniboxRenderer keyword arguments:
excluded     - a list of field names that shall not be rendered
forced       - a list of field names, excluded by default, that shall be rendered anyway
order        - a list of fields to align left (ordered left-to-right)
order_right  - a list of fields to align right (ordered left-to-right)
alignment    - field alignment (currently only "left" is supported)
substitute   - field name substitutions: { "real_name" : "rendered_name", ...}
individually - whether or not to render each object individually (default: False)
"""

class RendererBase(Renderer):
    EXCLUDED_FIELDS = ["id"]
    LEFT_ALIGNED_FIELDS = ["name"]
    RIGHT_ALIGNED_FIELDS = ["created", "created_at", "modified", "modified_at"]
    COLUMN_SPACING = 2
    ERROR_MISSING_FIELD = "<FIELD N/A>"
    def _render(self, obj):
        if raw_mode.is_on():
            # Normal rendering should not occur in raw mode
            return
        self._render_impl(obj)
    def _render_impl(self, obj):
        raise NotImplementedError
    def _render_field_names(self, field_names):
        """ :param field_names: field names in lower case """
        substitutions = { k.lower() : v.lower() for k, v in self._properties.get("substitute", {}).iteritems() }
        return [substitutions.get(field, field).upper().replace("_", " ") for field in field_names]
    def _get_ordered_and_rendered_fields(self, first_object):
        # Excluded fields
        forcefully_included_fields = self._properties.get("forced", ())
        excluded_fields = set(f for f in self.EXCLUDED_FIELDS if f not in forcefully_included_fields)
        excluded_fields.update(f.lower() for f in self._properties.get("excluded", ()))
        # Field ordering
        valid_fields = set(first_object).difference(excluded_fields)
        left_aligned_fields  = [f.lower() for f in self._properties.get("order",        self.LEFT_ALIGNED_FIELDS) if f in valid_fields]
        right_aligned_fields = [f.lower() for f in self._properties.get("order_right", self.RIGHT_ALIGNED_FIELDS) if f in valid_fields]
        unordered_fields = (f for f in valid_fields if f not in left_aligned_fields and f not in right_aligned_fields)
        # Field lengths
        all_fields = list(chain(left_aligned_fields, unordered_fields, right_aligned_fields))
        rendered_field_names = self._render_field_names(all_fields)
        return all_fields, rendered_field_names, excluded_fields

class InfiniboxRenderer(RendererBase):
    def __init__(self, *args, **kwargs):
        super(InfiniboxRenderer, self).__init__(*args, **kwargs)
        self._value_renderer = ValueRenderer()
    def _remove_redundant_fields(self, obj):
        status = obj.get("status", None)
        if isinstance(status, dict) and "status_code" in status and "status_description" in status:
            obj.pop("status")
        obj.pop("obj_name", None)
    def _render_impl(self, obj):
        self._render_root(obj)
    def _headerize(self, s):
        return s.title().replace("_", " ")
    def _render_root(self, obj):
        if obj is None:
            return self._render_null()
        elif isinstance(obj, PagedQuery):
            objects = obj.get_all_objects()
            return self._render_root(objects)
        elif hasattr(obj, "__len__") and len(obj) == 0:
            return self._render_empty()
        elif isinstance(obj, (list, set, tuple)):
            if len(obj) == 1 and isinstance(obj[0], (list, set, tuple, dict)):
                return self._render_root(obj[0])
            each_is_dict = map(lambda o: isinstance(o, dict), obj)
            if False in each_is_dict:
                if True in each_is_dict:
                    # Some items are dicts, and some aren't
                    return self._render_mixed_objects_list(obj)
                # All items are non-dicts
                return self._writeln(", ".join(unicode(o) for o in obj))
            # All items are dicts
            return self._render_similar_objects_list(obj)
        elif isinstance(obj, dict):
            self._remove_redundant_fields(obj)
            # If after stripping there's only one item in this dict,
            # start the rendering process over with that item
            if len(obj) == 1:
                # Render object
                return self._render_root(next(obj.itervalues()))
            else:
                if any(isinstance(o, dict) for o in obj.itervalues()):
                    return self._render_mixed_objects_dict(obj)
                else:
                    # Assume this dict represents one item, rather than being a container of items
                    return self._render_similar_objects_list((obj,))
        elif hasattr(obj, "__iter__"):
            return self._render_root(list(obj))
        else:
            return self._writeln(obj if isinstance(obj, basestring) else pformat(obj))
    def _render_mixed_objects_list(self, objects):
        for obj in objects:
            self._render_root(obj)
    def _render_mixed_objects_dict(self, objects):
        for key, value in objects.iteritems():
            header = self._headerize(key)
            if isinstance(value, (basestring, Number)):
                self._writeln(u"{} :  {}".format(header, value))
            else:
                self._writeln(header)
                with self._indent():
                    self._render_root(value)
    def _render_similar_objects_list(self, objects):
        if self._properties.get("individually", False):
            self._render_each_object_individually(objects)
        else:
            self._render_similar_objects_table(objects)
    def _render_each_object_individually(self, objects):
        for obj in objects:
            max_field_len = max(len(field) for field in obj)
            header_template = u"{{:<{}}}   ".format(max_field_len)
            for field, value in obj.iteritems():
                header = self._headerize(field)
                value = self._render_value(obj, field, value)
                self._write(header_template.format(header), colors=[Style.BRIGHT])
                self._writeln(value)
    def _get_fields(self, first_object):
        all_fields, rendered_field_names, excluded_fields = self._get_ordered_and_rendered_fields(first_object)
        fields = OrderedDict((field, len(rendered_field_names[i])) for i, field in enumerate(all_fields))
        return fields, rendered_field_names, excluded_fields
    def _render_similar_objects_table(self, objects):
        for obj in objects:
            self._remove_redundant_fields(obj)
        fields, rendered_field_names, excluded_fields = self._get_fields(objects[0])
        for o in objects:
            for field, value in o.items():
                if field in excluded_fields:
                    continue
                # Ensure all objects are of the same type, i.e. have the same fields
                if field not in fields:
                    # And if not, then render them differently
                    return self._render_mixed_objects_list(objects)
                # Render value
                rendered = self._render_value(o, field, value)
                if value is not rendered:
                    o[field] = rendered
                    value = rendered
                # Set max width for this field
                fields[field] = max(fields[field], len(unicode(value)))
            for field in fields:
                if field not in o:
                    o[field] = self.ERROR_MISSING_FIELD
                    fields[field] = max(len(self.ERROR_MISSING_FIELD), fields[field])
        self._render_table_rows(objects, fields, rendered_field_names)
    def _render_table_rows(self, objects, fields, rendered_field_names):
        formatter = u"  ".join(self._iter_formatter_hunks(fields))
        self._writeln(formatter.format(*(field for field in rendered_field_names)), colors=[Style.BRIGHT])
        for o in objects:
            # Format values, while keeping fields order
            self._writeln(formatter.format(*(o[field] for field in fields)))
    def _iter_default_alignments(self):
        return self._iter_left_alignments()
    def _iter_right_alignments(self):
        yield "<"
        while True:
            yield ">"
    def _iter_left_alignments(self):
        while True:
            yield "<"
    def _iter_formatter_hunks(self, fields):
        if not fields:
            return
        alignment = self._properties.get("alignment")
        if alignment is not None:
            if isinstance(alignment, basestring) and alignment.strip().lower() == "left":
                alignments = self._iter_left_alignments()
            else:
                raise NotImplementedError
        else:
            alignments = self._iter_default_alignments()
        widths = fields.itervalues()
        for align, width in izip(alignments, widths):
            yield u"{{:{0}{1}}}".format(align, width)
    def _render_value(self, obj, field, value):
        if self._is_field_rendered(obj, field):
            return value
        return self._value_renderer.render(field, value)
    def _is_field_rendered(self, obj, field):
        rendered_fields = obj.__dict__.get("__rendered_fields")
        if not rendered_fields:
            obj.__dict__["__rendered_fields"] = rendered_fields = set()
        if field in rendered_fields:
            return True
        # Mark field as rendered
        rendered_fields.add(field)
        return False
    def _render_empty(self):
        self._writeln("")
    def _render_null(self):
        pass

class InfiniboxShellRenderer(InfiniboxRenderer):
    def _render_empty(self):
        self._writeln("<empty>")

class CollectionRenderer(RendererBase):
    def __init__(self, *args, **kwargs):
        super(CollectionRenderer, self).__init__(*args, **kwargs)
        self._value_renderer = ValueRenderer()
    def _validate_result(self, obj):
        if isinstance(obj, PagedQuery) or hasattr(obj, "__iter__"):
            return
        raise UnexpectedAPIResponse("Result format is unexpected, and cannot be displayed")
    def _render_impl(self, objects):
        self._validate_result(objects)
        it = iter(objects)
        try:
            first_object = next(it)
        except StopIteration:
            return
        fields = self._get_fields(first_object)
        all_objects = chain([first_object], it)
        self._render_objects(all_objects, fields)
    def _get_fields(self, first_object):
        all_fields, rendered_field_names, _ = self._get_ordered_and_rendered_fields(first_object)
        return FieldSpecs.get(
                entity_type=self._properties.get("entity"),
                field_names=all_fields,
                rendered_field_names=rendered_field_names,
                console_width=None if not self._stream.isatty() else get_console_size()[0],
                column_spacing=self.COLUMN_SPACING,
                )
    def _iter_default_alignments(self):
        return self._iter_left_alignments()
    def _iter_right_alignments(self):
        yield "<"
        while True:
            yield ">"
    def _iter_left_alignments(self):
        while True:
            yield "<"
    def _iter_formatter_hunks(self, fields):
        if not fields:
            return
        alignment = self._properties.get("alignment")
        if alignment is not None:
            assert isinstance(alignment, basestring)
            alignment = alignment.strip().lower()
            if alignment == "left":
                alignments = self._iter_left_alignments()
            elif alignment == "right":
                alignments = self._iter_right_alignments()
            else:
                raise NotImplementedError
        else:
            alignments = self._iter_default_alignments()
        widths = (field.length for field in fields)
        for align, width in izip(alignments, widths):
            yield u"{{:{0}{1}}}".format(align, width)
    def _render_headers(self, template, fields):
        self._writeln(template.format(*(field.rendered_name for field in fields)), colors=[Style.BRIGHT])
    def _render_objects(self, objects, fields):
        template = (u" " * self.COLUMN_SPACING).join(self._iter_formatter_hunks(fields))
        self._render_headers(template, fields)
        for object in objects:
            values = (self._render_value(object, field) for field in fields)
            line = template.format(*values)
            self._writeln(line)
    def _render_value(self, object, field):
        value = object.get(field.name, self.ERROR_MISSING_FIELD)
        if value is not self.ERROR_MISSING_FIELD:
            value = self._value_renderer.render(field.name, value)
        if len(value) > field.length:
            value = value[:field.length-3] + '...'
        return value

class FieldSpec(object):
    def __init__(self, name, rendered_name, length):
        self.name = name
        self.rendered_name = rendered_name
        self.length = length
    def is_dynamic(self):
        return isinstance(self.length, Weight)
    def __repr__(self):
        return represent(self, 'name', 'rendered_name', 'length')

class Weight(object):
    def __init__(self, weight, fixed, min=None, max=None):
        '''
        :param weight: weight to be used if console width is available
        :param fixed: fixed length to be used if console width is unavailable
        :param min: minimal length (fixed/2 if unspecified)
        :param max: maximal length (fixed*2 if unspecified)
        '''
        self.weight = weight
        self.fixed = fixed
        self.min = min or (fixed / 2)
        self.max = max or 800 or (fixed * 2)
    def get_length(self, factor, field_name_length):
        exact_length = self.weight / factor
        rounded_length = int(round(exact_length))
        real_min = max(self.min, field_name_length)
        if rounded_length <= real_min:
            return real_min
        return rounded_length
    def get_value_length(self, factor):
        return self.get_length(factor, self.min)
    def __repr__(self):
        return represent(self, 'weight', 'fixed', 'min', 'max')

class FieldSpecs(object):
    # Presets
    ID = 10
    NAME = Weight(weight=10, fixed=20)
    DATE = len('0000-00-00 00:00:00')
    CAPACITY = len('0000.000 TB')
    # Fields
    ENTITY_FIELDS = {
            None : {
                'id' : ID,
                'name|description' : NAME,
                '(timestamp|created|modified)(_at)?' : DATE,
                '.*(size|capacity).*' : CAPACITY,
                'volume|host|cluster' : NAME,
                'pool' : Weight(weight=5, fixed=10)
                },
            types.Pool : {
                'disk_unused' : 16,
                },
            types.Volume : {
                'type' : 6,
                },
            types.Host : {
                },
            types.Cluster : {
                'hosts' : NAME,
                'luns' : NAME,
                },
            types.Event : {
                'description' : Weight(weight=25, fixed=100, max=400),
                'level' : len('CRITICAL'),
                'code' : 32,
                },
            }
    # Actual policy with precompiled patterns
    POLICY = { entity_type : tuple((re.compile('(' + field + ')$', re.I), length)
                                   for field, length in fields.iteritems())
               for entity_type, fields in ENTITY_FIELDS.iteritems() }
    @classmethod
    def get(cls, entity_type, field_names, rendered_field_names, console_width, column_spacing):
        assert len(field_names) == len(rendered_field_names)
        specs = [FieldSpec(
                        name=f,
                        rendered_name=rendered_field_names[i],
                        length=cls._get_field(entity_type, f, rendered_field_names[i]),
                        )
                 for i, f in enumerate(field_names)]
        if not console_width:
            # Console width is unknown - use fixed sizes
            for spec in specs:
                if isinstance(spec.length, Weight):
                    spec.length = max(spec.length.fixed, len(spec.rendered_name))
        else:
            # Console width is known - use dynamic sizing
            remaining_width = console_width - (len(specs) - 1) * column_spacing
            dynamic_fields = []
            dynamic_weight = 0.0 # keep this a float
            for spec in specs:
                if isinstance(spec.length, Weight):
                    dynamic_fields.append(spec)
                    dynamic_weight += spec.length.weight
                else:
                    remaining_width -= spec.length
            if remaining_width <= 0:
                # Fixed widths exceeds console width - use minimal width for dynamic fields
                for spec in dynamic_fields:
                    spec.length = max(spec.length.min, len(spec.rendered_name))
            else:
                # There's room for dynamic fields - compute their widths
                factor = 0
                while True:
                    field_removed = False
                    factor = dynamic_weight / remaining_width
                    for spec in dynamic_fields:
                        if spec.length.get_value_length(factor) > spec.length.max:
                            dynamic_fields.remove(spec)
                            field_removed = True
                            remaining_width -= spec.length.max
                            dynamic_weight -= spec.length.weight
                            spec.length = max(spec.length.max, len(spec.rendered_name))
                            break
                    if not field_removed:
                        break
                if dynamic_fields:
                    for spec in dynamic_fields:
                        field_length = spec.length.get_length(factor, len(spec.rendered_name))
                        spec.length = field_length
                        remaining_width -= field_length
                    if remaining_width > 0:
                        dynamic_fields[0].length += 1
                    if remaining_width < 0:
                        dynamic_fields[-1].length -= 1
        return specs
    @classmethod
    def _get_field(cls, entity_type, field_name, rendered_field_name):
        # Get specific patterns for entity
        policy = cls.POLICY.get(entity_type)
        is_default_policy = False
        if policy is None:
            policy = cls.POLICY.get(None)
            is_default_policy = True
        length = cls._match_field(policy, field_name)
        if length is None and not is_default_policy:
            length = cls._match_field(cls.POLICY.get(None), field_name)
        length = len(rendered_field_name) if length is None \
                 else max(length, len(rendered_field_name)) if isinstance(length, Number) \
                 else length
        return length
    @classmethod
    def _match_field(cls, policy, field_name):
        for pattern, length in policy:
            if pattern.match(field_name):
                return length
        return None
