from ..utils import PossiblyNone
import re
import datetime
import time
from numbers import Number

__all__ = ['ValueRenderer']

class ValueRenderer(object):
    RENDERERS = None # Value is set at module bottom
    def __init__(self, empty_value="-"):
        self._empty_value = empty_value
    def render(self, field, value):
        if isinstance(value, PossiblyNone) and not value._is_none:
            value = value._value
        if value is None or value == "" or (hasattr(value, "__len__") and not value):
            return self._empty_value
        try:
            for renderer in self.RENDERERS:
                if renderer.is_suitable(field, value):
                    return unicode(renderer.render(value))
        except Exception as e:
            return u"[{}]".format(e).upper()
        return unicode(value)

# === ABSTRACT VALUE RENDERERS ===
class ValueTypeRenderer(object):
    def is_suitable(self, field, value):
        raise NotImplementedError
    def render(self, obj):
        raise NotImplementedError
class PatternSearchValueTypeRenderer(ValueTypeRenderer):
    class __metaclass__(type):
        def __new__(mcls, name, bases, attrs):
            cls = type.__new__(mcls, name, bases, attrs)
            if isinstance(cls.PATTERN, basestring):
                cls.PATTERN = re.compile(cls.PATTERN, re.IGNORECASE)
            return cls
    PATTERN = None
    VALUE_TYPE = None
    def is_suitable(self, field, value):
        if self.VALUE_TYPE and not isinstance(value, self.VALUE_TYPE):
            return False
        return self.PATTERN.search(field) is not None
# === SOLID VALUE RENDERERS ===
class CompoundObjectRenderer(ValueTypeRenderer):
    def is_suitable(self, field, value):
        return isinstance(value, (dict, list, set, tuple))
    def render(self, obj):
        return "<compound>"
class StorageCapacityRenderer(PatternSearchValueTypeRenderer):
    PATTERN = "(SIZE|CAPACITY|SPACE|DISK_USAGE)"
    VALUE_TYPE = Number
    _capacities = ((float(1000**4), "TB"),
                   (float(1000**3), "GB"),
                   (float(1000**2), "MB"),
                   (float(1000), "KB"))
    def render(self, bytes):
        for i, (unit_bytes, unit) in enumerate(self._capacities):
            if bytes >= unit_bytes or round(bytes/(unit_bytes/1000), 2) >= 1000:
                return "{:.2f} {}".format(bytes/unit_bytes, unit)
        return bytes
class StorageCapacityTerabytesRenderer(StorageCapacityRenderer):
    def render(self, bytes):
        bytes = bytes or 0
        return "{:.3f} TB".format(bytes/(1000.0**4))
class MillisecondsDeltaRenderer(PatternSearchValueTypeRenderer):
    PATTERN = "UPTIME"
    VALUE_TYPE = Number
    def render(self, timestamp):
        timestamp = timestamp/1000.0
        now_seconds = time.time()
        now  = datetime.datetime.fromtimestamp(long(now_seconds))
        then = datetime.datetime.fromtimestamp(long(now_seconds - timestamp))
        return str(now - then)
class MillisecondsTimestampRenderer(PatternSearchValueTypeRenderer):
    PATTERN = "(TIME|DATE|(CREATED|MODIFIED)_AT)"
    VALUE_TYPE = Number
    def render(self, timestamp):
        if isinstance(timestamp, basestring):
            timestamp = long(timestamp)
        return datetime.datetime.fromtimestamp(timestamp/1000.0).strftime("%Y-%m-%d %H:%M:%S")
class SimpleArrayRenderer(ValueTypeRenderer):
    def is_suitable(self, field, value):
        if not isinstance(value, (list, set, tuple)):
            return False
        for v in value:
            if not isinstance(v, (Number, basestring)):
                return False
        return True
    def render(self, obj):
        return u", ".join(unicode(v) for v in obj)

# Order is important; the first suitable renderer is used
ValueRenderer.RENDERERS = (
    StorageCapacityTerabytesRenderer(),
    MillisecondsDeltaRenderer(),
    MillisecondsTimestampRenderer(),
    SimpleArrayRenderer(),
    CompoundObjectRenderer(),
    )

