from bunch import Bunch
from infi.pyutils.lazy import cached_property

from ..exceptions import Unauthorized, Forbidden, ManagementException
from ..commands.types import ConstantEntity
from .user import User

class System(object):
    def __init__(self, api):
        self.api = api
        self._user = None
    @cached_property
    def name(self):
        return self.api.get('system/name')
    def _login_and_get_user(self, username, password):
        with self.api.no_basic_auth_context():
            return self.api.post("users/login", dict(uid=username, password=password))
    def check_credentials(self, username, password):
        """ Attempt to login, but do not change the current user. Return True if
        login succeeded, else False. """
        try:
            self._login_and_get_user(username, password)
        except (Unauthorized, Forbidden):
            return False
        except ManagementException as e:
            if e.http_code in (401, 403):
                return False
            raise
        return True
    def login(self, username, password):
        user_attrs = self._login_and_get_user(username, password)
        self._set_user(**user_attrs)
        self.api.set_basic_auth(username, password)
    def _set_user(self, **user_attrs):
        self._user = User(Bunch(user_attrs))
        ConstantEntity.invalidate_all_caches()
    def logout(self):
        self._user = None
    def is_logged_in(self):
        return self._user is not None
    @property
    def user(self):
        assert self.is_logged_in(), 'User not set - invoke login(username, password)'
        return self._user
    def __enter__(self):
        push_system(self)
    def __exit__(self, exc_type, exc_instance, traceback):
        pop_system()



### System context ##

_systems = []

def get_system(safe=False):
    if not _systems:
        assert safe, 'System not set' # TODO: Need a more appropriate exception here
        return None
    return _systems[-1]
def push_system(system):
    _systems.append(system)
def pop_system():
    _systems.pop()


### Singleton ###

class _SystemProxy(object):
    def __getattr__(self, attr):
        return getattr(get_system(), attr)
    def __setattr__(self, attr, value):
        setattr(get_system(), attr, value)

system = _SystemProxy()

del _SystemProxy
